from sqlalchemy.types import Float

from . import idsorted
from ..model import Fruit, Tree


def test_get_tree_with_property(rest, http):
    rest(Tree, properties=['fruit_colors'])
    code, json = http.get('/api/tree')
    assert code == 200
    assert json['occurences'] == 3
    assert idsorted(json['objects']) == [
        {
            'id': 1,
            'name': 'pine',
            'fruit_colors': 'grey, darkgrey, brown'
        },
        {
            'id': 2,
            'name': 'maple',
            'fruit_colors': 'red, orangered'
        },
        {
            'id': 3,
            'name': 'oak',
            'fruit_colors': ''
        },
    ]


def test_get_tree_with_object_property(rest, http):
    rest(Tree, properties=[rest.Property('fruit_colors')])
    code, json = http.get('/api/tree')
    assert code == 200
    assert json['occurences'] == 3
    assert idsorted(json['objects']) == [
        {
            'id': 1,
            'name': 'pine',
            'fruit_colors': 'grey, darkgrey, brown'
        },
        {
            'id': 2,
            'name': 'maple',
            'fruit_colors': 'red, orangered'
        },
        {
            'id': 3,
            'name': 'oak',
            'fruit_colors': ''
        },
    ]


results = [{
    'fruit_id': 1,
    'color': 'grey',
    'size': 12.0,
    'age': 1041300.0,
    'tree_id': 1,
    'square_size': 144.0
}, {
    'fruit_id': 2,
    'color': 'darkgrey',
    'size': 23.0,
    'age': 4233830.213,
    'tree_id': 1,
    'square_size': 529.0
}, {
    'fruit_id': 3,
    'color': 'brown',
    'size': 2.12,
    'age': 0.0,
    'tree_id': 1,
    'square_size': 4.4944
}, {
    'fruit_id': 4,
    'color': 'red',
    'size': 0.5,
    'age': 2400.0,
    'tree_id': 2,
    'square_size': 0.25
}, {
    'fruit_id': 5,
    'color': 'orangered',
    'size': 100.0,
    'age': 7200.000012,
    'tree_id': 2,
    'square_size': 10000.0
}]


def test_get_fruit_with_hybrid_property_and_type(rest, http):
    rest(Fruit, properties=[rest.Property('square_size', Float())])
    code, json = http.get('/api/fruit')
    assert code == 200
    assert json['occurences'] == 5
    assert idsorted(json['objects'], 'fruit_id') == results


def test_get_fruit_with_hybrid_property_and_formatter(rest, http):
    rest(
        Fruit,
        properties=[
            rest.Property('square_size', formatter=lambda x: float(x))
        ]
    )
    code, json = http.get('/api/fruit')
    assert code == 200
    assert json['occurences'] == 5
    assert idsorted(json['objects'], 'fruit_id') == results


def test_get_fruit_with_primary_keys(rest, http):
    rest(Fruit, primary_keys=['fruit_id', 'age'], only=[])
    code, json = http.get('/api/fruit')
    assert code == 200
    assert json['occurences'] == 5
    assert json['primary_keys'] == ['fruit_id', 'age']
    assert idsorted(json['objects'], 'fruit_id') == [{
        'fruit_id': 1,
        'age': 1041300.0,
    }, {
        'fruit_id': 2,
        'age': 4233830.213,
    }, {
        'fruit_id': 3,
        'age': 0.0,
    }, {
        'fruit_id': 4,
        'age': 2400.0,
    }, {
        'fruit_id': 5,
        'age': 7200.000012,
    }]



from sqlalchemy.types import Float

from unrest import UnRest, __about__
from unrest.flask_framework import FlaskUnRest

from ..model import Fruit, Tree
from .openapi_result import openapi


def test_normal(app, db, http):
    rest = UnRest(app, db.session)
    rest(Tree)
    code, json = http.get('/api/tree')
    assert code == 200
    assert json['occurences'] == 3


def test_path(app, db, http):
    rest = UnRest(app, db.session, '/forest')
    rest(Tree)
    code, json = http.get('/forest/tree')
    assert code == 200
    assert json['occurences'] == 3


def test_version(app, db, http):
    rest = UnRest(app, db.session, version='v3.14')
    rest(Tree)
    code, json = http.get('/api/v3.14/tree')
    assert code == 200
    assert json['occurences'] == 3


def test_path_and_version(app, db, http):
    rest = UnRest(app, db.session, '/forest', 'v3.14')
    rest(Tree)
    code, json = http.get('/forest/v3.14/tree')
    assert code == 200
    assert json['occurences'] == 3


def test_explicit_framework(app, db, http):
    rest = UnRest(app, db.session, framework=FlaskUnRest)
    rest(Tree)
    code, json = http.get('/api/tree')
    assert code == 200
    assert json['occurences'] == 3


def test_empty_explicit_framework(app, db, http):
    class FakeUnRest(object):
        def __init__(self, app, prefix):
            pass

        def register_route(self, *args, **kwargs):
            pass

        request_json = register_route
        send_json = register_route
        send_error = register_route

    rest = UnRest(app, db.session, framework=FakeUnRest)
    rest(Tree)
    code, _ = http.get('/api/tree')
    assert code == 404


def test_api_options(app, db, http):
    rest = UnRest(app, db.session)
    fruit = rest(Fruit)
    rest(
        Tree,
        methods=rest.all,
        relationships={'fruits': fruit},
        properties=['fruit_colors'],
        allow_batch=True
    )
    code, json = http.options('/api')
    assert code == 200
    assert json == {
        '/api/fruit': {
            'model': 'Fruit',
            'description': 'A bag of fruit',
            'parameters': ['fruit_id'],
            'columns': {
                'age': 'timedelta',
                'color': 'str',
                'fruit_id': 'int',
                'size': 'Decimal',
                'tree_id': 'int'
            },
            'properties': {},
            'relationships': {},
            'methods': ['GET', 'OPTIONS'],
            'batch': False
        },
        '/api/tree': {
            'model': 'Tree',
            'description': "Where money doesn't grow",
            'parameters': ['id'],
            'columns': {
                'id': 'int',
                'name': 'str'
            },
            'properties': {
                'fruit_colors': 'str'
            },
            'relationships': {
                'fruits': {
                    'model': 'Fruit',
                    'description': 'A bag of fruit',
                    'parameters': ['fruit_id'],
                    'columns': {
                        'age': 'timedelta',
                        'color': 'str',
                        'fruit_id': 'int',
                        'size': 'Decimal',
                        'tree_id': 'int'
                    },
                    'properties': {},
                    'relationships': {},
                    'batch': False
                }
            },
            'methods': ['GET', 'PUT', 'POST', 'DELETE', 'PATCH', 'OPTIONS'],
            'batch': True
        }
    }


def test_endpoint_options(app, db, http):
    rest = UnRest(app, db.session, framework=FlaskUnRest)
    fruit = rest(Fruit)
    rest(
        Tree,
        methods=rest.all,
        relationships={'fruits': fruit},
        properties=['fruit_colors'],
        allow_batch=True
    )
    code, json = http.options('/api/fruit')
    assert code == 200
    assert json == {
        'model': 'Fruit',
        'description': 'A bag of fruit',
        'batch': False,
        'parameters': ['fruit_id'],
        'columns': {
            'age': 'timedelta',
            'color': 'str',
            'fruit_id': 'int',
            'size': 'Decimal',
            'tree_id': 'int'
        },
        'properties': {},
        'relationships': {},
        'methods': ['GET', 'OPTIONS']
    }


def test_openapi(app, db, http):
    rest = UnRest(
        app,
        db.session,
        info={
            'description':
                '''# Unrest demo
This is the demo of unrest api.
This api expose the `Tree` and `Fruit` entity Rest methods.
''',
            'contact': {
                'name': __about__.__author__,
                'url': __about__.__uri__,
                'email': __about__.__email__
            },
            'license': {
                'name': __about__.__license__
            }
        }
    )
    fruit = rest(
        Fruit,
        methods=rest.all,
        properties=[rest.Property('square_size', Float())]
    )
    rest(
        Tree,
        methods=rest.all,
        relationships={'fruits': fruit},
        properties=['fruit_colors'],
        allow_batch=True
    )

    code, json = http.get('/api/openapi.json')
    assert code == 200
    assert json == openapi


def test_sub(app, db, http):
    rest = UnRest(app, db.session)

    fruit = rest(
        Fruit,
        methods=rest.all,
        properties=[rest.Property('square_size', Float())]
    )
    tree = rest(
        Tree,
        methods=rest.all,
        relationships={'fruits': fruit},
        properties=['fruit_colors'],
        query=lambda q: q.filter(Tree.name != 'pine'),
        allow_batch=True
    )
    subtree = tree.sub(lambda q: q.filter(Tree.name != 'oak'))
    for key in ['unrest', 'Model', 'methods', 'only', 'exclude', 'properties',
                'relationships', 'allow_batch', 'auth', 'read_auth',
                'write_auth', 'validators', '_primary_keys', 'SerializeClass',
                'DeserializeClass']:
        assert getattr(subtree, key) == getattr(tree, key)
    assert subtree.name == 'subtree'

    code, json = http.get('/api/tree')
    assert code == 200
    assert json['occurences'] == 2

    code, json = http.get('/api/subtree')
    assert code == 200
    assert json['occurences'] == 1

    subtree = tree.sub(
        lambda q: q.filter(Tree.name != 'maple'), name='nomaple'
    )
    code, json = http.get('/api/nomaple')
    assert code == 200
    assert json['occurences'] == 1



openapi = {
    "openapi": "3.0.0",
    "info": {
        "title":
            "unrest.tests.flask.conftest unrest api",
        "version":
            "1.0",
        "description":
            "# Unrest demo\nThis is the demo of unrest api.\n" +
            "This api expose the `Tree` and `Fruit` entity Rest methods.\n",
        "contact": {
            "name": "Florian Mounier",
            "url": "https://github.com/Kozea/unrest",
            "email": "florian.mounier@kozea.fr"
        },
        "license": {
            "name": "GNU LGPL v3+"
        }
    },
    "servers": [{
        "url": "http://localhost/api/"
    }],
    "paths": {
        "/fruit": {
            "get": {
                "tags": ["Fruit"],
                "summary": "Retrieve all Fruit objects",
                "responses": {
                    "200": {
                        "description": "All Fruit objects",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "type": "object",
                                    "properties": {
                                        "primary_keys": {
                                            "type": "array",
                                            "items": {
                                                "type": "string"
                                            }
                                        },
                                        "occurences": {
                                            "type": "integer"
                                        },
                                        "objects": {
                                            "type": "array",
                                            "items": {
                                                "type": "object",
                                                "properties": {
                                                    "fruit_id": {
                                                        "type": "integer",
                                                        "format": "int64"
                                                    },
                                                    "color": {
                                                        "type": "string"
                                                    },
                                                    "age": {
                                                        "type": "number"
                                                    },
                                                    "size": {
                                                        "type": "number",
                                                        "format": "double"
                                                    },
                                                    "tree_id": {
                                                        "type": "integer",
                                                        "format": "int64"
                                                    },
                                                    "square_size": {
                                                        "type": "number",
                                                        "format": "double"
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            },
            "post": {
                "tags": ["Fruit"],
                "summary": "Add a new Fruit object",
                "responses": {
                    "200": {
                        "description": "The new Fruit object",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "type": "object",
                                    "properties": {
                                        "primary_keys": {
                                            "type": "array",
                                            "items": {
                                                "type": "string"
                                            }
                                        },
                                        "occurences": {
                                            "type": "integer"
                                        },
                                        "objects": {
                                            "type": "array",
                                            "items": {
                                                "type": "object",
                                                "properties": {
                                                    "fruit_id": {
                                                        "type": "integer",
                                                        "format": "int64"
                                                    },
                                                    "color": {
                                                        "type": "string"
                                                    },
                                                    "age": {
                                                        "type": "number"
                                                    },
                                                    "size": {
                                                        "type": "number",
                                                        "format": "double"
                                                    },
                                                    "tree_id": {
                                                        "type": "integer",
                                                        "format": "int64"
                                                    },
                                                    "square_size": {
                                                        "type": "number",
                                                        "format": "double"
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                },
                "requestBody": {
                    "description": "The new Fruit object to create",
                    "content": {
                        "application/json": {
                            "schema": {
                                "type": "object",
                                "properties": {
                                    "fruit_id": {
                                        "type": "integer",
                                        "format": "int64"
                                    },
                                    "color": {
                                        "type": "string"
                                    },
                                    "age": {
                                        "type": "number"
                                    },
                                    "size": {
                                        "type": "number",
                                        "format": "double"
                                    },
                                    "tree_id": {
                                        "type": "integer",
                                        "format": "int64"
                                    }
                                }
                            }
                        }
                    },
                    "required": True
                }
            },
            "options": {
                "tags": ["Fruit"],
                "summary": "Get info about the Fruit collection",
                "responses": {
                    "200": {
                        "description": "Info about the Fruit collection",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "type": "object",
                                    "properties": {
                                        "primary_keys": {
                                            "type": "array",
                                            "items": {
                                                "type": "string"
                                            }
                                        },
                                        "occurences": {
                                            "type": "integer"
                                        },
                                        "objects": {
                                            "type": "array",
                                            "items": {
                                                "type": "object",
                                                "properties": {
                                                    "fruit_id": {
                                                        "type": "integer",
                                                        "format": "int64"
                                                    },
                                                    "color": {
                                                        "type": "string"
                                                    },
                                                    "age": {
                                                        "type": "number"
                                                    },
                                                    "size": {
                                                        "type": "number",
                                                        "format": "double"
                                                    },
                                                    "tree_id": {
                                                        "type": "integer",
                                                        "format": "int64"
                                                    },
                                                    "square_size": {
                                                        "type": "number",
                                                        "format": "double"
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            },
            "summary": "A bag of fruit"
        },
        "/fruit/{fruit_id}": {
            "get": {
                "tags": ["Fruit"],
                "summary":
                    "Retrieve the corresponding Fruit object",
                "responses": {
                    "200": {
                        "description": "The corresponding Fruit object",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "type": "object",
                                    "properties": {
                                        "primary_keys": {
                                            "type": "array",
                                            "items": {
                                                "type": "string"
                                            }
                                        },
                                        "occurences": {
                                            "type": "integer"
                                        },
                                        "objects": {
                                            "type": "array",
                                            "items": {
                                                "type": "object",
                                                "properties": {
                                                    "fruit_id": {
                                                        "type": "integer",
                                                        "format": "int64"
                                                    },
                                                    "color": {
                                                        "type": "string"
                                                    },
                                                    "age": {
                                                        "type": "number"
                                                    },
                                                    "size": {
                                                        "type": "number",
                                                        "format": "double"
                                                    },
                                                    "tree_id": {
                                                        "type": "integer",
                                                        "format": "int64"
                                                    },
                                                    "square_size": {
                                                        "type": "number",
                                                        "format": "double"
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                },
                "parameters": [{
                    "name": "fruit_id",
                    "in": "path",
                    "required": True,
                    "schema": {
                        "type": "string"
                    }
                }]
            },
            "put": {
                "tags": ["Fruit"],
                "summary":
                    "Replace the corresponding Fruit object or create it",
                "responses": {
                    "200": {
                        "description": "The added Fruit object",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "type": "object",
                                    "properties": {
                                        "primary_keys": {
                                            "type": "array",
                                            "items": {
                                                "type": "string"
                                            }
                                        },
                                        "occurences": {
                                            "type": "integer"
                                        },
                                        "objects": {
                                            "type": "array",
                                            "items": {
                                                "type": "object",
                                                "properties": {
                                                    "fruit_id": {
                                                        "type": "integer",
                                                        "format": "int64"
                                                    },
                                                    "color": {
                                                        "type": "string"
                                                    },
                                                    "age": {
                                                        "type": "number"
                                                    },
                                                    "size": {
                                                        "type": "number",
                                                        "format": "double"
                                                    },
                                                    "tree_id": {
                                                        "type": "integer",
                                                        "format": "int64"
                                                    },
                                                    "square_size": {
                                                        "type": "number",
                                                        "format": "double"
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                },
                "parameters": [{
                    "name": "fruit_id",
                    "in": "path",
                    "required": True,
                    "schema": {
                        "type": "string"
                    }
                }],
                "requestBody": {
                    "description": "The Fruit object to create or replace",
                    "content": {
                        "application/json": {
                            "schema": {
                                "type": "object",
                                "properties": {
                                    "fruit_id": {
                                        "type": "integer",
                                        "format": "int64"
                                    },
                                    "color": {
                                        "type": "string"
                                    },
                                    "age": {
                                        "type": "number"
                                    },
                                    "size": {
                                        "type": "number",
                                        "format": "double"
                                    },
                                    "tree_id": {
                                        "type": "integer",
                                        "format": "int64"
                                    }
                                }
                            }
                        }
                    },
                    "required": True
                }
            },
            "delete": {
                "tags": ["Fruit"],
                "summary":
                    "Delete the corresponding Fruit object",
                "responses": {
                    "200": {
                        "description": "The deleted Fruit object",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "type": "object",
                                    "properties": {
                                        "primary_keys": {
                                            "type": "array",
                                            "items": {
                                                "type": "string"
                                            }
                                        },
                                        "occurences": {
                                            "type": "integer"
                                        },
                                        "objects": {
                                            "type": "array",
                                            "items": {
                                                "type": "object",
                                                "properties": {
                                                    "fruit_id": {
                                                        "type": "integer",
                                                        "format": "int64"
                                                    },
                                                    "color": {
                                                        "type": "string"
                                                    },
                                                    "age": {
                                                        "type": "number"
                                                    },
                                                    "size": {
                                                        "type": "number",
                                                        "format": "double"
                                                    },
                                                    "tree_id": {
                                                        "type": "integer",
                                                        "format": "int64"
                                                    },
                                                    "square_size": {
                                                        "type": "number",
                                                        "format": "double"
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                },
                "parameters": [{
                    "name": "fruit_id",
                    "in": "path",
                    "required": True,
                    "schema": {
                        "type": "string"
                    }
                }]
            },
            "patch": {
                "tags": ["Fruit"],
                "summary":
                    "Patch the corresponding Fruit object",
                "responses": {
                    "200": {
                        "description": "The patched Fruit object",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "type": "object",
                                    "properties": {
                                        "primary_keys": {
                                            "type": "array",
                                            "items": {
                                                "type": "string"
                                            }
                                        },
                                        "occurences": {
                                            "type": "integer"
                                        },
                                        "objects": {
                                            "type": "array",
                                            "items": {
                                                "type": "object",
                                                "properties": {
                                                    "fruit_id": {
                                                        "type": "integer",
                                                        "format": "int64"
                                                    },
                                                    "color": {
                                                        "type": "string"
                                                    },
                                                    "age": {
                                                        "type": "number"
                                                    },
                                                    "size": {
                                                        "type": "number",
                                                        "format": "double"
                                                    },
                                                    "tree_id": {
                                                        "type": "integer",
                                                        "format": "int64"
                                                    },
                                                    "square_size": {
                                                        "type": "number",
                                                        "format": "double"
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                },
                "parameters": [{
                    "name": "fruit_id",
                    "in": "path",
                    "required": True,
                    "schema": {
                        "type": "string"
                    }
                }],
                "requestBody": {
                    "description": "The Fruit object patches.",
                    "content": {
                        "application/json": {
                            "schema": {
                                "type": "object",
                                "properties": {
                                    "fruit_id": {
                                        "type": "integer",
                                        "format": "int64"
                                    },
                                    "color": {
                                        "type": "string"
                                    },
                                    "age": {
                                        "type": "number"
                                    },
                                    "size": {
                                        "type": "number",
                                        "format": "double"
                                    },
                                    "tree_id": {
                                        "type": "integer",
                                        "format": "int64"
                                    }
                                }
                            }
                        }
                    },
                    "required": True
                }
            },
            "options": {
                "tags": ["Fruit"],
                "summary":
                    "Get info about the Fruit collection",
                "responses": {
                    "200": {
                        "description": "Info about the Fruit collection",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "type": "object",
                                    "properties": {
                                        "primary_keys": {
                                            "type": "array",
                                            "items": {
                                                "type": "string"
                                            }
                                        },
                                        "occurences": {
                                            "type": "integer"
                                        },
                                        "objects": {
                                            "type": "array",
                                            "items": {
                                                "type": "object",
                                                "properties": {
                                                    "fruit_id": {
                                                        "type": "integer",
                                                        "format": "int64"
                                                    },
                                                    "color": {
                                                        "type": "string"
                                                    },
                                                    "age": {
                                                        "type": "number"
                                                    },
                                                    "size": {
                                                        "type": "number",
                                                        "format": "double"
                                                    },
                                                    "tree_id": {
                                                        "type": "integer",
                                                        "format": "int64"
                                                    },
                                                    "square_size": {
                                                        "type": "number",
                                                        "format": "double"
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                },
                "parameters": [{
                    "name": "fruit_id",
                    "in": "path",
                    "required": True,
                    "schema": {
                        "type": "string"
                    }
                }]
            },
            "summary": "A bag of fruit"
        },
        "/tree": {
            "get": {
                "tags": ["Tree"],
                "summary": "Retrieve all Tree objects",
                "responses": {
                    "200": {
                        "description": "All Tree objects",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "type": "object",
                                    "properties": {
                                        "primary_keys": {
                                            "type": "array",
                                            "items": {
                                                "type": "string"
                                            }
                                        },
                                        "occurences": {
                                            "type": "integer"
                                        },
                                        "objects": {
                                            "type": "array",
                                            "items": {
                                                "type": "object",
                                                "properties": {
                                                    "id": {
                                                        "type": "integer",
                                                        "format": "int64"
                                                    },
                                                    "name": {
                                                        "type": "string"
                                                    },
                                                    "fruit_colors": {
                                                        "type": "string"
                                                    },
                                                    "fruits": {
                                                        "type": "array",
                                                        "items": {
                                                            "type": "object",
                                                            "properties": {
                                                                "fruit_id": {
                                                                    "type":
                                                                        "integer",
                                                                    "format":
                                                                        "int64"
                                                                },
                                                                "color": {
                                                                    "type":
                                                                        "string"
                                                                },
                                                                "age": {
                                                                    "type":
                                                                        "number"
                                                                },
                                                                "size": {
                                                                    "type":
                                                                        "number",
                                                                    "format":
                                                                        "double"
                                                                },
                                                                "tree_id": {
                                                                    "type":
                                                                        "integer",
                                                                    "format":
                                                                        "int64"
                                                                },
                                                                "square_size": {
                                                                    "type":
                                                                        "number",
                                                                    "format":
                                                                        "double"
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            },
            "put": {
                "tags": ["Tree"],
                "summary": "Replace all Tree objects",
                "responses": {
                    "200": {
                        "description": "The new Tree objects",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "type": "object",
                                    "properties": {
                                        "primary_keys": {
                                            "type": "array",
                                            "items": {
                                                "type": "string"
                                            }
                                        },
                                        "occurences": {
                                            "type": "integer"
                                        },
                                        "objects": {
                                            "type": "array",
                                            "items": {
                                                "type": "object",
                                                "properties": {
                                                    "id": {
                                                        "type": "integer",
                                                        "format": "int64"
                                                    },
                                                    "name": {
                                                        "type": "string"
                                                    },
                                                    "fruit_colors": {
                                                        "type": "string"
                                                    },
                                                    "fruits": {
                                                        "type": "array",
                                                        "items": {
                                                            "type": "object",
                                                            "properties": {
                                                                "fruit_id": {
                                                                    "type":
                                                                        "integer",
                                                                    "format":
                                                                        "int64"
                                                                },
                                                                "color": {
                                                                    "type":
                                                                        "string"
                                                                },
                                                                "age": {
                                                                    "type":
                                                                        "number"
                                                                },
                                                                "size": {
                                                                    "type":
                                                                        "number",
                                                                    "format":
                                                                        "double"
                                                                },
                                                                "tree_id": {
                                                                    "type":
                                                                        "integer",
                                                                    "format":
                                                                        "int64"
                                                                },
                                                                "square_size": {
                                                                    "type":
                                                                        "number",
                                                                    "format":
                                                                        "double"
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                },
                "requestBody": {
                    "description":
                        "The new Tree objects to replace the current collection with",
                    "content": {
                        "application/json": {
                            "schema": {
                                "type": "object",
                                "properties": {
                                    "objects": {
                                        "type": "array",
                                        "items": {
                                            "type": "object",
                                            "properties": {
                                                "id": {
                                                    "type": "integer",
                                                    "format": "int64"
                                                },
                                                "name": {
                                                    "type": "string"
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    },
                    "required":
                        True
                }
            },
            "post": {
                "tags": ["Tree"],
                "summary": "Add a new Tree object",
                "responses": {
                    "200": {
                        "description": "The new Tree object",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "type": "object",
                                    "properties": {
                                        "primary_keys": {
                                            "type": "array",
                                            "items": {
                                                "type": "string"
                                            }
                                        },
                                        "occurences": {
                                            "type": "integer"
                                        },
                                        "objects": {
                                            "type": "array",
                                            "items": {
                                                "type": "object",
                                                "properties": {
                                                    "id": {
                                                        "type": "integer",
                                                        "format": "int64"
                                                    },
                                                    "name": {
                                                        "type": "string"
                                                    },
                                                    "fruit_colors": {
                                                        "type": "string"
                                                    },
                                                    "fruits": {
                                                        "type": "array",
                                                        "items": {
                                                            "type": "object",
                                                            "properties": {
                                                                "fruit_id": {
                                                                    "type":
                                                                        "integer",
                                                                    "format":
                                                                        "int64"
                                                                },
                                                                "color": {
                                                                    "type":
                                                                        "string"
                                                                },
                                                                "age": {
                                                                    "type":
                                                                        "number"
                                                                },
                                                                "size": {
                                                                    "type":
                                                                        "number",
                                                                    "format":
                                                                        "double"
                                                                },
                                                                "tree_id": {
                                                                    "type":
                                                                        "integer",
                                                                    "format":
                                                                        "int64"
                                                                },
                                                                "square_size": {
                                                                    "type":
                                                                        "number",
                                                                    "format":
                                                                        "double"
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                },
                "requestBody": {
                    "description": "The new Tree object to create",
                    "content": {
                        "application/json": {
                            "schema": {
                                "type": "object",
                                "properties": {
                                    "id": {
                                        "type": "integer",
                                        "format": "int64"
                                    },
                                    "name": {
                                        "type": "string"
                                    }
                                }
                            }
                        }
                    },
                    "required": True
                }
            },
            "delete": {
                "tags": ["Tree"],
                "summary": "Delete all Tree objects",
                "responses": {
                    "200": {
                        "description": "All deleted Tree objects",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "type": "object",
                                    "properties": {
                                        "primary_keys": {
                                            "type": "array",
                                            "items": {
                                                "type": "string"
                                            }
                                        },
                                        "occurences": {
                                            "type": "integer"
                                        },
                                        "objects": {
                                            "type": "array",
                                            "items": {
                                                "type": "object",
                                                "properties": {
                                                    "id": {
                                                        "type": "integer",
                                                        "format": "int64"
                                                    },
                                                    "name": {
                                                        "type": "string"
                                                    },
                                                    "fruit_colors": {
                                                        "type": "string"
                                                    },
                                                    "fruits": {
                                                        "type": "array",
                                                        "items": {
                                                            "type": "object",
                                                            "properties": {
                                                                "fruit_id": {
                                                                    "type":
                                                                        "integer",
                                                                    "format":
                                                                        "int64"
                                                                },
                                                                "color": {
                                                                    "type":
                                                                        "string"
                                                                },
                                                                "age": {
                                                                    "type":
                                                                        "number"
                                                                },
                                                                "size": {
                                                                    "type":
                                                                        "number",
                                                                    "format":
                                                                        "double"
                                                                },
                                                                "tree_id": {
                                                                    "type":
                                                                        "integer",
                                                                    "format":
                                                                        "int64"
                                                                },
                                                                "square_size": {
                                                                    "type":
                                                                        "number",
                                                                    "format":
                                                                        "double"
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            },
            "patch": {
                "tags": ["Tree"],
                "summary": "Patch all Tree objects",
                "responses": {
                    "200": {
                        "description": "The patched Tree objects",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "type": "object",
                                    "properties": {
                                        "primary_keys": {
                                            "type": "array",
                                            "items": {
                                                "type": "string"
                                            }
                                        },
                                        "occurences": {
                                            "type": "integer"
                                        },
                                        "objects": {
                                            "type": "array",
                                            "items": {
                                                "type": "object",
                                                "properties": {
                                                    "id": {
                                                        "type": "integer",
                                                        "format": "int64"
                                                    },
                                                    "name": {
                                                        "type": "string"
                                                    },
                                                    "fruit_colors": {
                                                        "type": "string"
                                                    },
                                                    "fruits": {
                                                        "type": "array",
                                                        "items": {
                                                            "type": "object",
                                                            "properties": {
                                                                "fruit_id": {
                                                                    "type":
                                                                        "integer",
                                                                    "format":
                                                                        "int64"
                                                                },
                                                                "color": {
                                                                    "type":
                                                                        "string"
                                                                },
                                                                "age": {
                                                                    "type":
                                                                        "number"
                                                                },
                                                                "size": {
                                                                    "type":
                                                                        "number",
                                                                    "format":
                                                                        "double"
                                                                },
                                                                "tree_id": {
                                                                    "type":
                                                                        "integer",
                                                                    "format":
                                                                        "int64"
                                                                },
                                                                "square_size": {
                                                                    "type":
                                                                        "number",
                                                                    "format":
                                                                        "double"
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                },
                "requestBody": {
                    "description": "The Tree objects patches",
                    "content": {
                        "application/json": {
                            "schema": {
                                "type": "object",
                                "properties": {
                                    "objects": {
                                        "type": "array",
                                        "items": {
                                            "type": "object",
                                            "properties": {
                                                "id": {
                                                    "type": "integer",
                                                    "format": "int64"
                                                },
                                                "name": {
                                                    "type": "string"
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    },
                    "required": True
                }
            },
            "options": {
                "tags": ["Tree"],
                "summary": "Get info about the Tree collection",
                "responses": {
                    "200": {
                        "description": "Info about the Tree collection",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "type": "object",
                                    "properties": {
                                        "primary_keys": {
                                            "type": "array",
                                            "items": {
                                                "type": "string"
                                            }
                                        },
                                        "occurences": {
                                            "type": "integer"
                                        },
                                        "objects": {
                                            "type": "array",
                                            "items": {
                                                "type": "object",
                                                "properties": {
                                                    "id": {
                                                        "type": "integer",
                                                        "format": "int64"
                                                    },
                                                    "name": {
                                                        "type": "string"
                                                    },
                                                    "fruit_colors": {
                                                        "type": "string"
                                                    },
                                                    "fruits": {
                                                        "type": "array",
                                                        "items": {
                                                            "type": "object",
                                                            "properties": {
                                                                "fruit_id": {
                                                                    "type":
                                                                        "integer",
                                                                    "format":
                                                                        "int64"
                                                                },
                                                                "color": {
                                                                    "type":
                                                                        "string"
                                                                },
                                                                "age": {
                                                                    "type":
                                                                        "number"
                                                                },
                                                                "size": {
                                                                    "type":
                                                                        "number",
                                                                    "format":
                                                                        "double"
                                                                },
                                                                "tree_id": {
                                                                    "type":
                                                                        "integer",
                                                                    "format":
                                                                        "int64"
                                                                },
                                                                "square_size": {
                                                                    "type":
                                                                        "number",
                                                                    "format":
                                                                        "double"
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            },
            "summary": "Where money doesn't grow"
        },
        "/tree/{id}": {
            "get": {
                "tags": ["Tree"],
                "summary": "Retrieve the corresponding Tree object",
                "responses": {
                    "200": {
                        "description": "The corresponding Tree object",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "type": "object",
                                    "properties": {
                                        "primary_keys": {
                                            "type": "array",
                                            "items": {
                                                "type": "string"
                                            }
                                        },
                                        "occurences": {
                                            "type": "integer"
                                        },
                                        "objects": {
                                            "type": "array",
                                            "items": {
                                                "type": "object",
                                                "properties": {
                                                    "id": {
                                                        "type": "integer",
                                                        "format": "int64"
                                                    },
                                                    "name": {
                                                        "type": "string"
                                                    },
                                                    "fruit_colors": {
                                                        "type": "string"
                                                    },
                                                    "fruits": {
                                                        "type": "array",
                                                        "items": {
                                                            "type": "object",
                                                            "properties": {
                                                                "fruit_id": {
                                                                    "type":
                                                                        "integer",
                                                                    "format":
                                                                        "int64"
                                                                },
                                                                "color": {
                                                                    "type":
                                                                        "string"
                                                                },
                                                                "age": {
                                                                    "type":
                                                                        "number"
                                                                },
                                                                "size": {
                                                                    "type":
                                                                        "number",
                                                                    "format":
                                                                        "double"
                                                                },
                                                                "tree_id": {
                                                                    "type":
                                                                        "integer",
                                                                    "format":
                                                                        "int64"
                                                                },
                                                                "square_size": {
                                                                    "type":
                                                                        "number",
                                                                    "format":
                                                                        "double"
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                },
                "parameters": [{
                    "name": "id",
                    "in": "path",
                    "required": True,
                    "schema": {
                        "type": "string"
                    }
                }]
            },
            "put": {
                "tags": ["Tree"],
                "summary":
                    "Replace the corresponding Tree object or create it",
                "responses": {
                    "200": {
                        "description": "The added Tree object",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "type": "object",
                                    "properties": {
                                        "primary_keys": {
                                            "type": "array",
                                            "items": {
                                                "type": "string"
                                            }
                                        },
                                        "occurences": {
                                            "type": "integer"
                                        },
                                        "objects": {
                                            "type": "array",
                                            "items": {
                                                "type": "object",
                                                "properties": {
                                                    "id": {
                                                        "type": "integer",
                                                        "format": "int64"
                                                    },
                                                    "name": {
                                                        "type": "string"
                                                    },
                                                    "fruit_colors": {
                                                        "type": "string"
                                                    },
                                                    "fruits": {
                                                        "type": "array",
                                                        "items": {
                                                            "type": "object",
                                                            "properties": {
                                                                "fruit_id": {
                                                                    "type":
                                                                        "integer",
                                                                    "format":
                                                                        "int64"
                                                                },
                                                                "color": {
                                                                    "type":
                                                                        "string"
                                                                },
                                                                "age": {
                                                                    "type":
                                                                        "number"
                                                                },
                                                                "size": {
                                                                    "type":
                                                                        "number",
                                                                    "format":
                                                                        "double"
                                                                },
                                                                "tree_id": {
                                                                    "type":
                                                                        "integer",
                                                                    "format":
                                                                        "int64"
                                                                },
                                                                "square_size": {
                                                                    "type":
                                                                        "number",
                                                                    "format":
                                                                        "double"
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                },
                "parameters": [{
                    "name": "id",
                    "in": "path",
                    "required": True,
                    "schema": {
                        "type": "string"
                    }
                }],
                "requestBody": {
                    "description": "The Tree object to create or replace",
                    "content": {
                        "application/json": {
                            "schema": {
                                "type": "object",
                                "properties": {
                                    "id": {
                                        "type": "integer",
                                        "format": "int64"
                                    },
                                    "name": {
                                        "type": "string"
                                    }
                                }
                            }
                        }
                    },
                    "required": True
                }
            },
            "delete": {
                "tags": ["Tree"],
                "summary": "Delete the corresponding Tree object",
                "responses": {
                    "200": {
                        "description": "The deleted Tree object",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "type": "object",
                                    "properties": {
                                        "primary_keys": {
                                            "type": "array",
                                            "items": {
                                                "type": "string"
                                            }
                                        },
                                        "occurences": {
                                            "type": "integer"
                                        },
                                        "objects": {
                                            "type": "array",
                                            "items": {
                                                "type": "object",
                                                "properties": {
                                                    "id": {
                                                        "type": "integer",
                                                        "format": "int64"
                                                    },
                                                    "name": {
                                                        "type": "string"
                                                    },
                                                    "fruit_colors": {
                                                        "type": "string"
                                                    },
                                                    "fruits": {
                                                        "type": "array",
                                                        "items": {
                                                            "type": "object",
                                                            "properties": {
                                                                "fruit_id": {
                                                                    "type":
                                                                        "integer",
                                                                    "format":
                                                                        "int64"
                                                                },
                                                                "color": {
                                                                    "type":
                                                                        "string"
                                                                },
                                                                "age": {
                                                                    "type":
                                                                        "number"
                                                                },
                                                                "size": {
                                                                    "type":
                                                                        "number",
                                                                    "format":
                                                                        "double"
                                                                },
                                                                "tree_id": {
                                                                    "type":
                                                                        "integer",
                                                                    "format":
                                                                        "int64"
                                                                },
                                                                "square_size": {
                                                                    "type":
                                                                        "number",
                                                                    "format":
                                                                        "double"
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                },
                "parameters": [{
                    "name": "id",
                    "in": "path",
                    "required": True,
                    "schema": {
                        "type": "string"
                    }
                }]
            },
            "patch": {
                "tags": ["Tree"],
                "summary":
                    "Patch the corresponding Tree object",
                "responses": {
                    "200": {
                        "description": "The patched Tree object",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "type": "object",
                                    "properties": {
                                        "primary_keys": {
                                            "type": "array",
                                            "items": {
                                                "type": "string"
                                            }
                                        },
                                        "occurences": {
                                            "type": "integer"
                                        },
                                        "objects": {
                                            "type": "array",
                                            "items": {
                                                "type": "object",
                                                "properties": {
                                                    "id": {
                                                        "type": "integer",
                                                        "format": "int64"
                                                    },
                                                    "name": {
                                                        "type": "string"
                                                    },
                                                    "fruit_colors": {
                                                        "type": "string"
                                                    },
                                                    "fruits": {
                                                        "type": "array",
                                                        "items": {
                                                            "type": "object",
                                                            "properties": {
                                                                "fruit_id": {
                                                                    "type":
                                                                        "integer",
                                                                    "format":
                                                                        "int64"
                                                                },
                                                                "color": {
                                                                    "type":
                                                                        "string"
                                                                },
                                                                "age": {
                                                                    "type":
                                                                        "number"
                                                                },
                                                                "size": {
                                                                    "type":
                                                                        "number",
                                                                    "format":
                                                                        "double"
                                                                },
                                                                "tree_id": {
                                                                    "type":
                                                                        "integer",
                                                                    "format":
                                                                        "int64"
                                                                },
                                                                "square_size": {
                                                                    "type":
                                                                        "number",
                                                                    "format":
                                                                        "double"
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                },
                "parameters": [{
                    "name": "id",
                    "in": "path",
                    "required": True,
                    "schema": {
                        "type": "string"
                    }
                }],
                "requestBody": {
                    "description": "The Tree object patches.",
                    "content": {
                        "application/json": {
                            "schema": {
                                "type": "object",
                                "properties": {
                                    "id": {
                                        "type": "integer",
                                        "format": "int64"
                                    },
                                    "name": {
                                        "type": "string"
                                    }
                                }
                            }
                        }
                    },
                    "required": True
                }
            },
            "options": {
                "tags": ["Tree"],
                "summary": "Get info about the Tree collection",
                "responses": {
                    "200": {
                        "description": "Info about the Tree collection",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "type": "object",
                                    "properties": {
                                        "primary_keys": {
                                            "type": "array",
                                            "items": {
                                                "type": "string"
                                            }
                                        },
                                        "occurences": {
                                            "type": "integer"
                                        },
                                        "objects": {
                                            "type": "array",
                                            "items": {
                                                "type": "object",
                                                "properties": {
                                                    "id": {
                                                        "type": "integer",
                                                        "format": "int64"
                                                    },
                                                    "name": {
                                                        "type": "string"
                                                    },
                                                    "fruit_colors": {
                                                        "type": "string"
                                                    },
                                                    "fruits": {
                                                        "type": "array",
                                                        "items": {
                                                            "type": "object",
                                                            "properties": {
                                                                "fruit_id": {
                                                                    "type":
                                                                        "integer",
                                                                    "format":
                                                                        "int64"
                                                                },
                                                                "color": {
                                                                    "type":
                                                                        "string"
                                                                },
                                                                "age": {
                                                                    "type":
                                                                        "number"
                                                                },
                                                                "size": {
                                                                    "type":
                                                                        "number",
                                                                    "format":
                                                                        "double"
                                                                },
                                                                "tree_id": {
                                                                    "type":
                                                                        "integer",
                                                                    "format":
                                                                        "int64"
                                                                },
                                                                "square_size": {
                                                                    "type":
                                                                        "number",
                                                                    "format":
                                                                        "double"
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                },
                "parameters": [{
                    "name": "id",
                    "in": "path",
                    "required": True,
                    "schema": {
                        "type": "string"
                    }
                }]
            },
            "summary": "Where money doesn't grow"
        }
    }
}



import json
import logging
from collections import OrderedDict
from copy import deepcopy
from functools import wraps

from sqlalchemy import and_, or_
from sqlalchemy.inspection import inspect
from sqlalchemy.orm.query import Query
from sqlalchemy.orm.strategy_options import Load
from sqlalchemy.schema import Column

from .coercers import Deserialize, Serialize
from .generators.options import Options

try:
    from json import JSONDecodeError
except ImportError:
    JSONDecodeError = Exception

log = logging.getLogger('unrest.rest')


class Rest(object):
    """
    This is the entry point for generating a REST endpoint for a specific model
    The final uri if the path is '/api' and version 'v2' would be:
    `/api/v2/model` and `/api/v2/model/pk1/pk2` and if model is not in the
    public schema `/api/v2/schema/model` and `/api/v2/schema/model/pk1/pk2`.

    Usage:
    ```python
        rest = UnRest(app)

        def name_validator(field):
            if len(field.value) > 12:
                raise field.ValidationError(
                    'Name is too long (max 12 characters).')
            return field.value

        rest(Person, only=['name', 'sex', 'age'], methods=rest.all,
             query=lambda q: q.filter(Person.age > 16),
             validators={'name': name_validator})
    ```
    # Arguments
        unrest: The unrest instance given automatically on UnRest call.
        Model: The sqlalchemy orm model class.
        methods: The allowed method list on this endpoint. Possible values are
            GET, PUT, POST, DELETE, PATCH and rest.all
        name: If specified replaces the model name in url.
        only: If specified restricts the json fields to this list.
        exclude: If specified removes the json fields in this list.
        query: A function that takes the Model query and returns your specific
            query. Can be useful to filter data for all the methods.
        properties: A list of additional properties to retrieve on the model.
        relationships: A mapping of relationships and rest endpoints to fetch
            with the model.
        allow_batch: Allow batch operations (PUT, DELETE and PATCH)
            without primary key.
        auth: A decorator that will always be called.
        read_auth: A decorator that will be called on GET.
        write_auth: A decorator that will be called on PUT, POST, DELETE
            and PATCH.
        validators: A mapping of field names and validation functions.
            A validator function takes a `Rest.Validatable` object as parameter
            and must return the final value for the field or raise a
            `rest.ValidationError(reason)` (where `rest = Unrest()`)
        primary_keys: A list of column names to use as primary_keys
            (use real db primary keys by default)
        defaults: A mapping of column -> values which sets the default value
            of the columns when the column is not present in the payload.
            fixed: A mapping of column -> values which replaces the values
            present or not in the payload.
        SerializeClass: An alternative #Serialize class.
        DeserializeClass: An alternative #Deserialize class.
    """

    def __init__(
            self,
            unrest,
            Model,
            methods=['GET'],
            name=None,
            only=None,
            exclude=None,
            query=None,
            properties=None,
            relationships=None,
            allow_batch=False,
            auth=None,
            read_auth=None,
            write_auth=None,
            validators=None,
            primary_keys=None,
            defaults=None,
            fixed=None,
            SerializeClass=Serialize,
            DeserializeClass=Deserialize
    ):
        self.unrest = unrest
        self.unrest.rests.append(self)
        self.Model = Model

        self.methods = methods[:]
        self.name = name or self.table.name
        self.only = only
        self.exclude = exclude
        self.query_factory = query or (lambda q: q)
        self.properties = [
            self.unrest.Property(property)
            if not isinstance(property, self.unrest.Property) else property
            for property in (properties or [])
        ]
        self.relationships = relationships or {}

        self.allow_batch = allow_batch

        self.auth = auth
        self.read_auth = read_auth
        self.write_auth = write_auth

        self.validators = validators or {}
        self._primary_keys = primary_keys
        self.defaults = defaults or {}
        self.fixed = fixed or {}

        self.SerializeClass = SerializeClass
        self.DeserializeClass = DeserializeClass

        if (self.unrest.allow_options and self.methods
                and 'OPTIONS' not in self.methods):
            self.methods.append('OPTIONS')

        for method in self.methods:
            self.register_method(method)

    def get(self, payload, **pks):
        """
        The GET method

        No arguments: Returns all query elements. (/api/model/)
        Primary keys: Returns the element in query with the primary keys or
            404. (/api/model/pk)

        # Arguments
            payload: The json request content ignored for GET.
            pks: The primary keys in url if any.
        """
        if self.has(pks):
            item = self.get_from_pk(self.query, **pks)
            if item is None:
                return self.unrest.Response(
                    self.serialize([]), status_code=404
                )

            return self.serialize([item])

        items = self.query
        return self.serialize(items)

    def put(self, payload, **pks):
        """
        The PUT method

        No arguments: If allow_batch set to true replace all the query elements
            with the ones in the request payload.
        Primary keys: Create or replace the element associated
            with the primary keys from the one in the request payload.

        # Arguments
            payload: The json request content containing new elements.
            pks: The primary keys in url if any.
        """
        if not payload:
            self.raise_error(400, 'You must provide a payload')

        if self.has(pks):
            for pk, val in pks.items():
                if pk in payload:
                    assert payload[pk] == val, (
                        'Incoherent primary_key (%s) in payload (%r) '
                        'and url (%r) for PUT' % (pk, payload[pk], val)
                    )
                else:
                    payload[pk] = val
            existingItem = self.get_from_pk(self.query, **pks)
            previousItem = deepcopy(existingItem)
            item = self.deserialize(payload, existingItem or self.Model())
            self.validate(item, previousItem)
            if existingItem is None:
                self.session.add(item)
            self.session.flush()
            return self.serialize([item])

        if not self.allow_batch:
            raise self.unrest.RestError(
                406, 'You must set allow_batch to True '
                'if you want to use batch methods.'
            )

        self.query.delete()
        items = self.deserialize_all(payload)
        self.validate_all(items)
        self.session.add_all(items)
        self.session.flush()
        return self.serialize(items)

    def post(self, payload, **pks):
        """
        The POST method

        No arguments: Add element from request payload.
        Primary keys: Correspond to new collection creation. Unused.

        # Arguments
            payload: The json request content containing the new element.
            pks: The primary keys in url if any.
        """
        if self.has(pks):
            # Create a collection?
            raise self.unrest.RestError(
                501, "POST with primary keys corresponds to collection "
                "creation. It's not implemented by default. "
                "If you want to update an item use the PUT method instead"
            )

        if not payload:
            self.raise_error(400, 'You must provide a payload')
        item = self.deserialize(payload, self.Model())
        self.session.add(item)
        self.validate(item)
        self.session.flush()
        return self.serialize([item])

    def delete(self, payload, **pks):
        """
        The DELETE method

        No arguments: If allow_batch set to true delete all query elements.
        Primary keys: Delete the element associated with the primary keys.

        # Arguments
            payload: The json request content ignored in DELETE.
            pks: The primary keys of the element to delete.
        """
        if self.has(pks):
            item = self.get_from_pk(self.undefered_query, **pks)
            if item is None:
                self.raise_error(404, '%s(%r) not found' % (self.name, pks))

            self.session.delete(item)
            self.session.flush()
            return self.serialize([item])

        if not self.allow_batch:
            raise self.unrest.RestError(
                406, 'You must set allow_batch to True '
                'if you want to use batch methods.'
            )

        items = self.undefered_query.all()
        self.query.delete()
        self.session.flush()
        return self.serialize(items)

    def patch(self, payload, **pks):
        """
        The PATCH method

        No arguments: If allow_batch set to true patch existing elements
            with element attributes specified in the request payload.
        Primary keys: Patch only one

        # Arguments
            payload: The json request content containing
                a list of attributes to be patched.
            pks: The primary keys of the element to patch.
        """
        if not payload:
            self.raise_error(400, 'You must provide a payload')

        if self.has(pks):
            for pk, val in pks.items():
                if pk in payload:
                    assert payload[pk] == val, (
                        'Incoherent primary_key (%s) in payload (%r) '
                        'and url (%r) for PATCH' % (pk, payload[pk], val)
                    )
                else:
                    payload[pk] = val
            item = self.get_from_pk(self.query, **pks)
            if item is None:
                self.raise_error(404, '%s(%r) not found' % (self.name, pks))
            self.deserialize(payload, item, blank_missing=False)
            self.validate(item)
            self.session.flush()
            return self.serialize([item])

        if not self.allow_batch:
            raise self.unrest.RestError(
                406, 'You must set allow_batch to True '
                'if you want to use batch methods.'
            )

        patches = payload['objects']
        # Get all concerned items
        items = self.get_all_from_pks(
            self.query, [{pk: patch[pk]
                          for pk in self.primary_keys} for patch in patches]
        )
        if len(items) < len(patches):
            for patch in patches:
                if len([it for pk in self.primary_keys for it in items
                        if getattr(it, pk) == patch[pk]]) == 0:
                    self.raise_error(
                        404, '%s(%r) not found' % (
                            self.name, {
                                key: val
                                for key, val in patch.items()
                                if key in self.primary_keys
                            }
                        )
                    )
        for patch in patches:
            # Get the patch item
            item = [
                it for pk in self.primary_keys for it in items
                if getattr(it, pk) == patch[pk]
            ][0]
            # Merge only patched colmuns
            self.deserialize(patch, item, blank_missing=False)
        self.validate_all(items)
        self.session.flush()
        return self.serialize(items)

    def options(self, payload, **pks):
        """
        The OPTIONS method

        Returns a description of this rest endpoint.
        """
        return Options(self.unrest).get_route(self)

    def declare(self, method, manual_commit=False):
        """
        A decorator to register an alternative method.
        The original is still callable with rest.{method}

        ```python
        fruit = rest(Fruit)

        @fruit.declare('GET')
        def get(payload, fruit_id=None):
            rv = fruit.get(payload, fruit_id=fruit_id)
            return {
                'occurences': rv['occurences'],
                'objects': [
                    {'id': obj['fruit_id']} for obj in rv['objects']
                ]
            }
        ```

        # Arguments
            method: The method to override ('GET' for exemple)
            manual_commit: Don't auto commit after the method.
        """

        def register_fun(fun):
            if self.unrest.allow_options and not self.methods:
                self.register_method('OPTIONS')
            self.register_method(method, fun, manual_commit)

        return register_fun

    def sub(self, query_factory, **kwargs):
        """This methods return a copy of the current rest endpoint and takes a
        {query_factory} argument to alter the current query.

        # Arguments
            query_factory: A function that takes the original query
                in parameter and returns a new query.
            **kwargs: Can be used to override Rest constructor arguments
                (query is not supported)
        """

        assert not kwargs.get('query'), 'query is not supported on sub rest'
        inherited = {
            'methods': self.methods,
            'name': 'sub' + self.name,
            'only': self.only,
            'exclude': self.exclude,
            'properties': self.properties,
            'relationships': self.relationships,
            'allow_batch': self.allow_batch,
            'auth': self.auth,
            'read_auth': self.read_auth,
            'write_auth': self.write_auth,
            'validators': self.validators,
            'primary_keys': self._primary_keys,
            'SerializeClass': self.SerializeClass,
            'DeserializeClass': self.DeserializeClass,
        }
        inherited.update(kwargs)
        subrest = self.__class__(self.unrest, self.Model, **inherited)
        subrest.query_factory = lambda q: query_factory(self.query_factory(q))
        return subrest

    def kwargs_to_pks(self, kwargs):
        if not kwargs:
            return {}
        item = self.Model()
        self.DeserializeClass(kwargs, self.primary_keys).merge(item)
        return {name: getattr(item, name) for name in self.primary_keys}

    def deserialize(self, payload, item, blank_missing=True):
        if blank_missing:
            columns = self.columns
        else:
            # Mind only provided columns
            columns = {
                name: column
                for name, column in self.columns.items() if name in payload
            }
        self.set_defaults(payload, columns)
        return self.DeserializeClass(payload, columns).merge(item)

    def deserialize_all(self, payload):
        for item in payload['objects']:
            self.set_defaults(item, self.columns)
        return self.DeserializeClass(payload, self.columns).create(self.Model)

    def serialize_object(self, item):
        return self.SerializeClass(
            item, self.columns, self.properties, self.relationships
        ).dict()

    def serialize(self, items):
        rv = {}
        rv['primary_keys'] = list(self.primary_keys.keys())

        if isinstance(items, Query):
            rv['occurences'] = items.offset(None).limit(None).count()
            if items._offset is not None:
                rv['offset'] = items._offset
            if items._limit is not None:
                rv['limit'] = items._limit

        rv['objects'] = [self.serialize_object(item) for item in items]
        if 'occurences' not in rv:
            rv['occurences'] = len(rv['objects'])
        return rv

    def set_defaults(self, payload, columns):
        for name, column in columns.items():
            if name in self.fixed:
                payload[name] = self.fixed[name]
            elif name not in payload and name in self.defaults:
                payload[name] = self.defaults[name]

    class Validatable(object):
        def __init__(self, value, name, previous, next, ValidationError):
            self.value = value
            self.name = name
            self.previous = previous
            self.next = next
            self.ValidationError = ValidationError

    def validate(self, item, existing=None, errors=None):
        """
        Validates all validators colums against validators.
        Raise RestError if validation errors.
        """
        with self.session.no_autoflush:
            valid = True
            error = {'fields': {}}
            for pk in self.primary_keys:
                error[pk] = getattr(item, pk)

            for key, validators in self.validators.items():
                field_errors = []
                if callable(validators):
                    validators = (validators, )
                try:
                    for validator in validators:
                        setattr(
                            item, key,
                            validator(
                                self.Validatable(
                                    getattr(item, key), key, existing, item,
                                    self.unrest.ValidationError
                                )
                            )
                        )
                except self.unrest.ValidationError as e:
                    valid = False
                    field_errors.append(e.message)
                if field_errors:
                    error['fields'][key] = '\n'.join(field_errors)
            if errors is not None:
                errors.append(error)
            elif not valid:
                self.raise_error(
                    500, 'Validation Error', extra={
                        'errors': [error]
                    }
                )

    def validate_all(self, items):
        errors = []
        for item in items:
            self.validate(item, errors=errors)
        if any(len(error['fields']) for error in errors):
            self.raise_error(500, 'Validation Error', extra={'errors': errors})

    def raise_error(self, status, message, extra=None):
        self.unrest.raise_error(status, message, extra)

    def wrap_native(self, method, method_fun, manual_commit=False):
        @wraps(method_fun)
        def wrapped(**kwargs):
            pks = self.kwargs_to_pks(kwargs)
            json = self.unrest.framework.request_json()
            try:
                try:
                    payload = self.unjson(json)
                except JSONDecodeError as e:
                    self.raise_error(400, 'JSON Error in payload: %s' % e)
                decorated = method_fun
                if method == 'GET' and self.read_auth:
                    decorated = self.read_auth(decorated)
                if (method in ['PUT', 'POST', 'DELETE', 'PATCH']
                        and self.write_auth):
                    decorated = self.write_auth(decorated)
                if self.auth:
                    decorated = self.auth(decorated)

                response = decorated(payload, **pks)

                if not manual_commit and method in ['PUT', 'POST', 'DELETE',
                                                    'PATCH']:
                    self.session.commit()
            except self.unrest.RestError as e:
                return self.unrest.framework.send_error(
                    dict(message=e.message, **e.extra), e.status
                )
            if not isinstance(response, self.unrest.Response):
                response = self.unrest.Response(response)
            log.info(
                '%s %s%s' % (
                    method, self.path,
                    ': %d occurences' % response.data['occurences']
                    if response.data.get('occurences') is not None else ''
                )
            )

            json = self.json(response.data)
            return response.wrapper(
                self.unrest.framework.send_json(json, response.status_code)
            )

        return wrapped

    def register_method(self, method, method_fun=None, manual_commit=False):
        if method != 'OPTIONS':
            assert method in self.unrest.all, 'Unknown method %s' % method
        method_fun = method_fun or getattr(self, method.lower())
        method_fun = self.wrap_native(method, method_fun, manual_commit)
        # str() for python 2 compat
        method_fun.__name__ = str('_'.join((method, ) + self.name_parts))
        self.unrest.framework.register_route(
            self.path, method, self.primary_keys, method_fun
        )

    def json(self, data):
        return json.dumps(data)

    def unjson(self, data):
        if data:
            return json.loads(data)

    def has(self, pks):
        return pks and all(val is not None for val in pks.values())

    def get_from_pk(self, query, **pks):
        for key, val in pks.items():
            query = query.filter(getattr(self.Model, key) == val)
        return query.first()

    def get_all_from_pks(self, query, items_pks):
        return query.filter(
            or_(
                *[
                    and_(
                        *[
                            getattr(self.Model, key) == val
                            for key, val in pks.items()
                        ]
                    ) for pks in items_pks
                ]
            )
        ).all()

    @property
    def session(self):
        return self.unrest.session

    @property
    def query(self):
        if hasattr(self.Model, 'query') and isinstance(self.Model.query, Query
                                                       ):
            query = self.Model.query
        else:
            query = self.session.query(self.Model)
        return self.query_factory(query)

    @property
    def undefered_query(self):
        return self.query.options(Load(self.Model).undefer('*'))

    @property
    def name_parts(self):
        if self.table.schema:
            return (self.table.schema, self.name)
        return (self.name, )

    @property
    def path(self):
        return '/'.join((self.unrest.root_path, ) + self.name_parts)

    @property
    def table(self):
        return self.Model.__table__

    @property
    def primary_keys(self):
        if self._primary_keys:
            return OrderedDict((name, column)
                               for name, column in self.model_columns
                               if name in self._primary_keys)
        ins = inspect(self.Model)
        return OrderedDict((ins.get_property_by_column(pk).key, pk)
                           for pk in ins.primary_key)

    @property
    def model_columns(self):
        for name, column in inspect(self.Model).columns.items():
            if isinstance(column, Column):
                yield name, column

    @property
    def columns(self):
        def gen():
            for name, column in self.model_columns:
                if name not in self.primary_keys:
                    if self.only is not None and name not in self.only:
                        continue
                    if self.exclude and name in self.exclude:
                        continue
                yield name, column

        return dict(gen())



from . import idsorted
from ..model import Fruit, Tree


def test_get_tree(rest, http):
    rest(Tree)
    rest(Fruit)
    code, json = http.get('/api/tree')
    assert code == 200
    assert json['occurences'] == 3
    assert idsorted(json['objects']) == [
        {
            'id': 1,
            'name': 'pine'
        },
        {
            'id': 2,
            'name': 'maple'
        },
        {
            'id': 3,
            'name': 'oak'
        },
    ]


def test_get_tree_name(rest, http):
    rest(Tree, name='forest')
    code, json = http.get('/api/forest')
    assert code == 200
    assert json['occurences'] == 3
    assert idsorted(json['objects']) == [
        {
            'id': 1,
            'name': 'pine'
        },
        {
            'id': 2,
            'name': 'maple'
        },
        {
            'id': 3,
            'name': 'oak'
        },
    ]


def test_get_tree_query(rest, http, db):
    def base_query(q):
        return db.session.query(Tree).filter(Tree.id > 1)

    rest(Tree, query=base_query)
    code, json = http.get('/api/tree')
    assert code == 200
    assert json['occurences'] == 2
    assert idsorted(json['objects']) == [
        {
            'id': 2,
            'name': 'maple'
        },
        {
            'id': 3,
            'name': 'oak'
        },
    ]


def test_get_tree_query_factory(rest, http, db):
    rest(Tree, query=lambda q: q.filter(Tree.id < 2))
    code, json = http.get('/api/tree')
    assert code == 200
    assert json['occurences'] == 1
    assert json['primary_keys'] == ['id']
    assert idsorted(json['objects']) == [{'id': 1, 'name': 'pine'}]


def test_get_fruits(rest, http):
    rest(Tree)
    rest(Fruit)
    code, json = http.get('/api/fruit')
    assert code == 200
    assert json['occurences'] == 5
    assert json['primary_keys'] == ['fruit_id']
    assert idsorted(json['objects'], 'fruit_id') == [{
        'fruit_id': 1,
        'color': 'grey',
        'size': 12.0,
        'age': 1041300.0,
        'tree_id': 1
    }, {
        'fruit_id': 2,
        'color': 'darkgrey',
        'size': 23.0,
        'age': 4233830.213,
        'tree_id': 1
    }, {
        'fruit_id': 3,
        'color': 'brown',
        'size': 2.12,
        'age': 0.0,
        'tree_id': 1
    }, {
        'fruit_id': 4,
        'color': 'red',
        'size': 0.5,
        'age': 2400.0,
        'tree_id': 2
    }, {
        'fruit_id': 5,
        'color': 'orangered',
        'size': 100.0,
        'age': 7200.000012,
        'tree_id': 2
    }]


def test_get_fruits_only(rest, http):
    rest(Fruit, only=['color', 'size'])
    code, json = http.get('/api/fruit')
    assert code == 200
    assert json['occurences'] == 5
    assert idsorted(json['objects'], 'fruit_id') == [{
        'fruit_id': 1,
        'color': 'grey',
        'size': 12.0
    }, {
        'fruit_id': 2,
        'color': 'darkgrey',
        'size': 23.0
    }, {
        'fruit_id': 3,
        'color': 'brown',
        'size': 2.12
    }, {
        'fruit_id': 4,
        'color': 'red',
        'size': 0.5
    }, {
        'fruit_id': 5,
        'color': 'orangered',
        'size': 100.0
    }]


def test_get_fruits_exclude(rest, http):
    rest(Fruit, exclude=['color', 'age'])
    code, json = http.get('/api/fruit')
    assert code == 200
    assert json['occurences'] == 5
    assert idsorted(json['objects'], 'fruit_id') == [{
        'fruit_id': 1,
        'size': 12.0,
        'tree_id': 1
    }, {
        'fruit_id': 2,
        'size': 23.0,
        'tree_id': 1
    }, {
        'fruit_id': 3,
        'size': 2.12,
        'tree_id': 1
    }, {
        'fruit_id': 4,
        'size': 0.5,
        'tree_id': 2
    }, {
        'fruit_id': 5,
        'size': 100.0,
        'tree_id': 2
    }]


def test_get_fruits_only_exclude(rest, http):
    rest(Fruit, only=['color', 'size'], exclude=['color', 'age'])
    code, json = http.get('/api/fruit')
    assert code == 200
    assert json['occurences'] == 5
    assert idsorted(json['objects'], 'fruit_id') == [{
        'fruit_id': 1,
        'size': 12.0
    }, {
        'fruit_id': 2,
        'size': 23.0
    }, {
        'fruit_id': 3,
        'size': 2.12
    }, {
        'fruit_id': 4,
        'size': 0.5
    }, {
        'fruit_id': 5,
        'size': 100.0
    }]


def test_no_method(rest, http):
    rest(Fruit, methods=[])
    code, json = http.get('/api/fruit')
    assert code == 404


def test_get_custom(rest, http):
    fruit = rest(Fruit)

    @fruit.declare('GET')
    def get(payload, fruit_id=None):
        return {'Hey': 'Overridden'}

    code, json = http.get('/api/fruit')
    assert code == 200
    assert json == {'Hey': 'Overridden'}


def test_get_custom_extend(rest, http):
    fruit = rest(Fruit)

    @fruit.declare('GET')
    def get(payload, fruit_id=None):
        rv = fruit.get(payload, fruit_id=fruit_id)
        return {
            'occurences': rv['occurences'],
            'objects': [{
                'id': obj['fruit_id']
            } for obj in rv['objects']]
        }

    code, json = http.get('/api/fruit')
    assert code == 200
    assert json['occurences'] == 5
    assert idsorted(json['objects']) == [{
        'id': 1
    }, {
        'id': 2
    }, {
        'id': 3
    }, {
        'id': 4
    }, {
        'id': 5
    }]



from tempfile import NamedTemporaryFile

from flask import Flask
from flask_sqlalchemy import SQLAlchemy
from sqlalchemy.types import Float

from unrest import UnRest, __about__

from ..model import Base, Fruit, Tree, fill_data

# Init flask
app = Flask(__name__)
f = NamedTemporaryFile()
app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False
app.config['SQLALCHEMY_DATABASE_URI'] = 'sqlite:///%s' % f.name


# Enable CORS
@app.after_request
def after_request(response):
    response.headers.add('Access-Control-Allow-Origin', '*')
    response.headers.add(
        'Access-Control-Allow-Headers', 'Content-Type,Authorization'
    )
    response.headers.add(
        'Access-Control-Allow-Methods', 'GET,PUT,POST,DELETE,PATCH,OPTIONS'
    )
    return response


# Init db
db = SQLAlchemy(app)
Base.metadata.drop_all(bind=db.engine)
Base.metadata.create_all(bind=db.engine)
fill_data(db.session)
db.session.remove()

# Optional info mapping (useful to enrich openapi file)
info = {
    'description':
        '''# Unrest demo
This is the demo of unrest api.
This api expose the `Tree` and `Fruit` entity Rest methods.
''',
    'contact': {
        'name': __about__.__author__,
        'url': __about__.__uri__,
        'email': __about__.__email__
    },
    'license': {
        'name': __about__.__license__
    }
}

# Init Unrest
rest = UnRest(app, db.session, info=info)
fruit = rest(
    Fruit,
    methods=rest.all,
    properties=[rest.Property('square_size', Float())]
)
rest(
    Tree,
    methods=rest.all,
    relationships={'fruits': fruit},
    properties=['fruit_colors'],
    allow_batch=True
)



from . import idsorted
from ..model import Tree


def test_paginated(rest, http):
    def paginated(query_factory, step=2):
        from flask import request

        def limited_query(query):
            if query_factory is not None:
                query = query_factory(query)
            offset = int(request.values.get('offset', 0))
            return query.offset(offset).limit(step)

        return limited_query

    rest(Tree, query=paginated(lambda query: query))
    code, json = http.get('/api/tree')
    assert code == 200
    assert json['occurences'] == 3
    assert json['limit'] == 2
    assert json['offset'] == 0
    assert json['primary_keys'] == ['id']
    assert idsorted(json['objects']) == [
        {
            'id': 1,
            'name': 'pine'
        },
        {
            'id': 2,
            'name': 'maple'
        },
    ]
    code, json = http.get('/api/tree?offset=1')
    assert code == 200
    assert json['occurences'] == 3
    assert json['limit'] == 2
    assert json['offset'] == 1
    assert json['primary_keys'] == ['id']
    assert idsorted(json['objects']) == [
        {
            'id': 2,
            'name': 'maple'
        },
        {
            'id': 3,
            'name': 'oak'
        },
    ]
    code, json = http.get('/api/tree?offset=2')
    assert code == 200
    assert json['occurences'] == 3
    assert json['limit'] == 2
    assert json['offset'] == 2
    assert json['primary_keys'] == ['id']
    assert idsorted(json['objects']) == [
        {
            'id': 3,
            'name': 'oak'
        },
    ]
    code, json = http.get('/api/tree?offset=3')
    assert code == 200
    assert json['occurences'] == 3
    assert json['limit'] == 2
    assert json['offset'] == 3
    assert json['primary_keys'] == ['id']
    assert idsorted(json['objects']) == []



from . import idsorted
from ..model import Fruit, Tree


def test_post_tree(rest, http):
    rest(Tree, methods=['GET', 'POST'])
    code, json = http.post('/api/tree', json={'name': 'cedar'})
    assert code == 200
    assert json['occurences'] == 1
    assert idsorted(json['objects']) == [
        {
            'id': 4,
            'name': 'cedar'
        },
    ]

    code, json = http.get('/api/tree')
    assert code == 200
    assert json['occurences'] == 4
    assert idsorted(json['objects']) == [
        {
            'id': 1,
            'name': 'pine'
        },
        {
            'id': 2,
            'name': 'maple'
        },
        {
            'id': 3,
            'name': 'oak'
        },
        {
            'id': 4,
            'name': 'cedar'
        },
    ]


def test_post_tree_with_id(rest, http):
    rest(Tree, methods=['GET', 'POST'])
    code, json = http.post('/api/tree', json={'id': 9, 'name': 'cedar'})
    assert code == 200
    assert json['occurences'] == 1
    assert idsorted(json['objects']) == [
        {
            'id': 9,
            'name': 'cedar'
        },
    ]

    code, json = http.get('/api/tree')
    assert code == 200
    assert json['occurences'] == 4
    assert idsorted(json['objects']) == [
        {
            'id': 1,
            'name': 'pine'
        },
        {
            'id': 2,
            'name': 'maple'
        },
        {
            'id': 3,
            'name': 'oak'
        },
        {
            'id': 9,
            'name': 'cedar'
        },
    ]


def test_post_tree_custom(rest, http):
    tree = rest(Tree)

    @tree.declare('POST')
    def post(payload, id=None):
        payload['name'] = 'I ALWAYS WANT THIS NAME'
        return tree.post(payload, id=id)

    code, json = http.post('/api/tree', json={'name': 'cedar'})
    assert code == 200
    assert json['occurences'] == 1
    assert idsorted(json['objects']) == [
        {
            'id': 4,
            'name': 'I ALWAYS WANT THIS NAME'
        },
    ]

    code, json = http.get('/api/tree')
    assert code == 200
    assert json['occurences'] == 4
    assert idsorted(json['objects']) == [
        {
            'id': 1,
            'name': 'pine'
        },
        {
            'id': 2,
            'name': 'maple'
        },
        {
            'id': 3,
            'name': 'oak'
        },
        {
            'id': 4,
            'name': 'I ALWAYS WANT THIS NAME'
        },
    ]


def test_post_tree_custom_manual_commit(rest, http):
    tree = rest(Tree)

    @tree.declare('POST', True)
    def post(payload, id=None):
        payload['name'] = 'I ALWAYS WANT THIS NAME'
        return tree.post(payload, id=id)

    code, json = http.post('/api/tree', json={'name': 'cedar'})
    assert code == 200
    assert json['occurences'] == 1
    assert idsorted(json['objects']) == [
        {
            'id': 4,
            'name': 'I ALWAYS WANT THIS NAME'
        },
    ]
    # The post should not be commited
    code, json = http.get('/api/tree')
    assert code == 200
    assert json['occurences'] == 3
    assert idsorted(json['objects']) == [{
        'id': 1,
        'name': 'pine'
    }, {
        'id': 2,
        'name': 'maple'
    }, {
        'id': 3,
        'name': 'oak'
    }]


def test_post_with_defaults(rest, http):
    rest(
        Fruit,
        methods=['GET', 'POST'],
        defaults={
            'color': 'white',
            'age': 42.0,
        }
    )
    code, json = http.post('/api/fruit', json={'size': 1.0, 'tree_id': 1})
    assert code == 200
    assert json['occurences'] == 1
    assert idsorted(json['objects'], 'fruit_id') == [
        {
            'fruit_id': 6,
            'color': 'white',
            'size': 1.0,
            'age': 42.0,
            'tree_id': 1
        },
    ]
    code, json = http.post(
        '/api/fruit', json={
            'color': 'yellow',
            'size': 2.0,
            'tree_id': 2
        }
    )
    assert code == 200
    assert json['occurences'] == 1
    assert idsorted(json['objects'], 'fruit_id') == [
        {
            'fruit_id': 7,
            'color': 'yellow',
            'size': 2.0,
            'age': 42.0,
            'tree_id': 2
        },
    ]

    code, json = http.get('/api/fruit')
    assert code == 200
    assert json['occurences'] == 7
    assert idsorted(json['objects'][-2:], 'fruit_id') == [
        {
            'fruit_id': 6,
            'color': 'white',
            'size': 1.0,
            'age': 42.0,
            'tree_id': 1
        },
        {
            'fruit_id': 7,
            'color': 'yellow',
            'size': 2.0,
            'age': 42.0,
            'tree_id': 2
        },
    ]


def test_post_with_fixed(rest, http):
    rest(
        Fruit,
        methods=['GET', 'POST'],
        fixed={
            'color': 'white',
            'age': 42.0,
        }
    )
    code, json = http.post('/api/fruit', json={'size': 1.0, 'tree_id': 1})
    assert code == 200
    assert json['occurences'] == 1
    assert idsorted(json['objects'], 'fruit_id') == [
        {
            'fruit_id': 6,
            'color': 'white',
            'size': 1.0,
            'age': 42.0,
            'tree_id': 1
        },
    ]
    code, json = http.post(
        '/api/fruit', json={
            'color': 'yellow',
            'size': 2.0,
            'tree_id': 2
        }
    )
    assert code == 200
    assert json['occurences'] == 1
    assert idsorted(json['objects'], 'fruit_id') == [
        {
            'fruit_id': 7,
            'color': 'white',
            'size': 2.0,
            'age': 42.0,
            'tree_id': 2
        },
    ]

    code, json = http.get('/api/fruit')
    assert code == 200
    assert json['occurences'] == 7
    assert idsorted(json['objects'][-2:], 'fruit_id') == [
        {
            'fruit_id': 6,
            'color': 'white',
            'size': 1.0,
            'age': 42.0,
            'tree_id': 1
        },
        {
            'fruit_id': 7,
            'color': 'white',
            'size': 2.0,
            'age': 42.0,
            'tree_id': 2
        },
    ]



from . import idsorted
from ..model import Fruit, Tree


def test_delete_tree(rest, http):
    rest(Tree, methods=['GET', 'DELETE'])
    code, json = http.delete('/api/tree/2')
    assert code == 200
    assert json['occurences'] == 1
    assert idsorted(json['objects']) == [
        {
            'id': 2,
            'name': 'maple'
        },
    ]

    code, json = http.get('/api/tree')
    assert code == 200
    assert json['occurences'] == 2
    assert idsorted(json['objects']) == [
        {
            'id': 1,
            'name': 'pine'
        },
        {
            'id': 3,
            'name': 'oak'
        },
    ]


def test_delete_unknown_tree(rest, http):
    rest(Tree, methods=['DELETE'])
    code, json = http.delete('/api/tree/9')
    assert code == 404


def test_delete_fruit(rest, http):
    rest(Fruit, methods=['GET', 'DELETE'])
    code, json = http.delete('/api/fruit/2')
    assert code == 200
    assert json['occurences'] == 1
    assert idsorted(json['objects'], 'fruit_id') == [
        {
            'fruit_id': 2,
            'color': 'darkgrey',
            'size': 23.0,
            'age': 4233830.213,
            'tree_id': 1
        },
    ]

    code, json = http.get('/api/fruit')
    assert code == 200
    assert json['occurences'] == 4
    assert idsorted(json['objects'], 'fruit_id') == [{
        'fruit_id': 1,
        'color': 'grey',
        'size': 12.0,
        'age': 1041300.0,
        'tree_id': 1
    }, {
        'fruit_id': 3,
        'color': 'brown',
        'size': 2.12,
        'age': 0.0,
        'tree_id': 1
    }, {
        'fruit_id': 4,
        'color': 'red',
        'size': 0.5,
        'age': 2400.0,
        'tree_id': 2
    }, {
        'fruit_id': 5,
        'color': 'orangered',
        'size': 100.0,
        'age': 7200.000012,
        'tree_id': 2
    }]



class Options(object):
    def __init__(self, unrest):
        self.unrest = unrest

    def get_property(self, type):
        try:
            return type.python_type.__name__
        except NotImplementedError:
            return type.__class__.__name__

    def get_columns(self, rest):
        return {
            name: self.get_property(column.type)
            for name, column in rest.columns.items()
        }

    def get_properties(self, rest):
        return {
            property.name: self.get_property(property.sqlalchemy_type)
            for property in rest.properties
        }

    def get_relationships(self, rest):
        return {
            name: {
                k: v
                for k, v in self.get_route(relationship).items()
                if k != 'methods'
            }
            for name, relationship in rest.relationships.items()
        }

    def get_route(self, rest):
        return {
            'model': rest.Model.__name__,
            'description': getattr(rest.Model, '__doc__', ''),
            'parameters': list(rest.primary_keys),
            'columns': self.get_columns(rest),
            'properties': self.get_properties(rest),
            'relationships': self.get_relationships(rest),
            'methods': rest.methods,
            'batch': rest.allow_batch
        }

    def all(self):
        return {rest.path: self.get_route(rest) for rest in self.unrest.rests}



__title__ = "unrest"
__version__ = "0.7.2"

__summary__ = "A troubling rest api library for sqlalchemy models"
__uri__ = "https://github.com/Kozea/unrest"
__author__ = "Florian Mounier"
__email__ = "florian.mounier@kozea.fr"

__license__ = "GNU LGPL v3+"
__copyright__ = "Copyright 2017 %s" % __author__



import json
from tempfile import NamedTemporaryFile

import pytest
from flask import Flask
from flask.testing import FlaskClient
from flask_sqlalchemy import SQLAlchemy

from unrest import UnRest

from ..model import Base, fill_data


class HTTPClient(FlaskClient):
    def open(self, *args, **kwargs):
        json_data = kwargs.pop('json', '')
        if json_data:
            kwargs.setdefault('data', json.dumps(json_data))
            kwargs.setdefault('content_type', 'application/json')

        response = super(HTTPClient, self).open(*args, **kwargs)

        if response.content_type == 'application/json':
            rv = json.loads(response.data.decode('utf-8'))
        else:
            rv = {'html': response.data.decode('utf-8')}
        return response.status_code, rv


@pytest.fixture
def app(request):
    app = Flask(__name__)
    app.test_client_class = HTTPClient
    f = NamedTemporaryFile()
    app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False
    app.config['SQLALCHEMY_DATABASE_URI'] = 'sqlite:///%s' % f.name
    return app


@pytest.fixture
def http(app):
    return app.test_client()


@pytest.fixture
def db_init(app):
    db = SQLAlchemy(app)
    Base.metadata.drop_all(bind=db.engine)
    return db


@pytest.fixture
def db(request, db_init):
    db = db_init
    Base.metadata.create_all(bind=db.engine)

    fill_data(db.session)
    db.session.remove()

    def drop_all():
        Base.metadata.drop_all(bind=db.engine)

    request.addfinalizer(drop_all)
    return db


@pytest.fixture
def rest(app, db):
    return UnRest(app, db.session)



from datetime import timedelta

from . import idsorted
from ..model import Fruit, Tree


def test_put_tree_validation(rest, http):
    def name_validator(field):
        if len(field.value) > 3:
            raise rest.ValidationError(
                'Name must be shorter than 4 characters.'
            )
        return field.value

    rest(
        Tree,
        methods=['GET', 'PUT'],
        validators={'name': name_validator},
        allow_batch=True
    )
    code, json = http.put(
        '/api/tree',
        json={
            'objects': [{
                'id': 1,
                'name': 'cedar'
            }, {
                'id': 2,
                'name': 'mango'
            }]
        }
    )
    assert code == 500
    assert json['message'] == 'Validation Error'
    assert json['errors'] == [{
        'id': 1,
        'fields': {
            'name': 'Name must be shorter than 4 characters.'
        },
    }, {
        'id': 2,
        'fields': {
            'name': 'Name must be shorter than 4 characters.'
        },
    }]
    code, json = http.get('/api/tree')
    assert code == 200
    assert json['occurences'] == 3
    assert idsorted(json['objects']) == [
        {
            'id': 1,
            'name': 'pine'
        },
        {
            'id': 2,
            'name': 'maple'
        },
        {
            'id': 3,
            'name': 'oak'
        },
    ]


def test_put_pk_tree_validation(rest, http):
    def name_validator(field):
        if len(field.value) > 3:
            raise rest.ValidationError(
                'Name must be shorter than 4 characters.'
            )
        return field.value

    rest(Tree, methods=['GET', 'PUT'], validators={'name': name_validator})
    code, json = http.put('/api/tree/1', json={'id': 1, 'name': 'cedar'})
    assert code == 500
    assert json['message'] == 'Validation Error'
    assert json['errors'] == [{
        'id': 1,
        'fields': {
            'name': 'Name must be shorter than 4 characters.'
        },
    }]
    code, json = http.get('/api/tree')
    assert code == 200
    assert json['occurences'] == 3
    assert idsorted(json['objects']) == [
        {
            'id': 1,
            'name': 'pine'
        },
        {
            'id': 2,
            'name': 'maple'
        },
        {
            'id': 3,
            'name': 'oak'
        },
    ]


def test_put_pk_tree_validation_ok(rest, http):
    def name_validator(field):
        if len(field.value) > 3:
            raise rest.ValidationError(
                'Name must be shorter than 4 characters.'
            )
        return field.value

    rest(Tree, methods=['GET', 'PUT'], validators={'name': name_validator})
    code, json = http.put('/api/tree/1', json={'id': 1, 'name': 'elm'})
    assert json['occurences'] == 1
    assert idsorted(json['objects']) == [
        {
            'id': 1,
            'name': 'elm'
        },
    ]

    code, json = http.get('/api/tree')
    assert code == 200
    assert json['occurences'] == 3
    assert idsorted(json['objects']) == [
        {
            'id': 1,
            'name': 'elm'
        },
        {
            'id': 2,
            'name': 'maple'
        },
        {
            'id': 3,
            'name': 'oak'
        },
    ]


def test_post_tree_validation(rest, http):
    def name_validator(field):
        if len(field.value) > 3:
            raise rest.ValidationError(
                'Name must be shorter than 4 characters.'
            )
        return field.value

    rest(Tree, methods=['GET', 'POST'], validators={'name': name_validator})
    code, json = http.post('/api/tree', json={'name': 'mango'})
    assert code == 500
    assert json['message'] == 'Validation Error'
    assert json['errors'] == [{
        'id': None,
        'fields': {
            'name': 'Name must be shorter than 4 characters.'
        },
    }]

    code, json = http.get('/api/tree')
    assert code == 200
    assert json['occurences'] == 3
    assert idsorted(json['objects']) == [
        {
            'id': 1,
            'name': 'pine'
        },
        {
            'id': 2,
            'name': 'maple'
        },
        {
            'id': 3,
            'name': 'oak'
        },
    ]


def test_put_fruit_dual_validation(rest, http):
    def color_validator(field):
        if field.previous.color == 'brown':
            raise field.ValidationError(
                'A brown fruit cannot become %s' % field.value
            )
        return field.value

    def size_validator(field):
        if field.value > 48.9 and field.next.age < timedelta(weeks=2):
            raise rest.ValidationError('Fruit too big for its age')
        return field.value

    def age_validator(field):
        if field.value < timedelta(days=100):
            raise field.ValidationError('Fruit too old')
        return field.value

    def tree_id_validator(field):
        if field.value == 0:
            raise rest.ValidationError('Invalid tree_id')
        return field.value

    rest(
        Fruit,
        methods=['GET', 'PUT'],
        validators={
            'color': color_validator,
            'size': size_validator,
            'age': age_validator,
            'tree_id': tree_id_validator
        }
    )
    code, json = http.put(
        '/api/fruit/3',
        json={
            'fruit_id': 3,
            'color': 'green',
            'size': 50.2,
            'age': 0,
            'tree_id': 0
        }
    )
    assert code == 500
    assert json['message'] == 'Validation Error'
    assert json['errors'] == [{
        'fruit_id': 3,
        'fields': {
            'color': 'A brown fruit cannot become green',
            'size': 'Fruit too big for its age',
            'age': 'Fruit too old',
            'tree_id': 'Invalid tree_id',
        },
    }]



from . import idsorted
from ..model import Fruit, Tree


def test_put_tree_implicitly_unallowed(rest, http):
    rest(Tree, methods=['GET', 'PUT'])
    code, json = http.put('/api/tree', json={'id': 1, 'name': 'cedar'})
    assert code == 406


def test_put_tree(rest, http):
    rest(Tree, methods=['GET', 'PUT'], allow_batch=True)
    code, json = http.put(
        '/api/tree',
        json={
            'objects': [{
                'id': 1,
                'name': 'cedar'
            }, {
                'id': 2,
                'name': 'mango'
            }]
        }
    )
    assert code == 200
    assert json['occurences'] == 2
    assert idsorted(json['objects']) == [
        {
            'id': 1,
            'name': 'cedar'
        },
        {
            'id': 2,
            'name': 'mango'
        },
    ]

    code, json = http.get('/api/tree')
    assert code == 200
    assert json['occurences'] == 2
    assert idsorted(json['objects']) == [
        {
            'id': 1,
            'name': 'cedar'
        },
        {
            'id': 2,
            'name': 'mango'
        },
    ]


def test_put_with_defaults(rest, http):
    rest(
        Fruit,
        methods=['GET', 'PUT'],
        allow_batch=True,
        defaults={
            'color': 'white',
            'age': 42.0,
        }
    )
    code, json = http.put(
        '/api/fruit',
        json={
            'objects': [{
                'size': 1.0,
                'tree_id': 1
            }, {
                'color': 'yellow',
                'size': 2.0,
                'tree_id': 2
            }]
        }
    )
    assert code == 200
    assert json['occurences'] == 2
    assert idsorted(json['objects'], 'fruit_id') == [
        {
            'fruit_id': 1,
            'color': 'white',
            'size': 1.0,
            'age': 42.0,
            'tree_id': 1
        },
        {
            'fruit_id': 2,
            'color': 'yellow',
            'size': 2.0,
            'age': 42.0,
            'tree_id': 2
        },
    ]

    code, json = http.get('/api/fruit')
    assert code == 200
    assert json['occurences'] == 2
    assert idsorted(json['objects'], 'fruit_id') == [
        {
            'fruit_id': 1,
            'color': 'white',
            'size': 1.0,
            'age': 42.0,
            'tree_id': 1
        },
        {
            'fruit_id': 2,
            'color': 'yellow',
            'size': 2.0,
            'age': 42.0,
            'tree_id': 2
        },
    ]


def test_put_with_fixed(rest, http):
    rest(
        Fruit,
        methods=['GET', 'PUT'],
        allow_batch=True,
        fixed={
            'color': 'white',
            'age': 42.0,
        }
    )
    code, json = http.put(
        '/api/fruit',
        json={
            'objects': [{
                'size': 1.0,
                'tree_id': 1
            }, {
                'color': 'yellow',
                'size': 2.0,
                'tree_id': 2
            }]
        }
    )
    assert code == 200
    assert json['occurences'] == 2
    assert idsorted(json['objects'], 'fruit_id') == [
        {
            'fruit_id': 1,
            'color': 'white',
            'size': 1.0,
            'age': 42.0,
            'tree_id': 1
        },
        {
            'fruit_id': 2,
            'color': 'white',
            'size': 2.0,
            'age': 42.0,
            'tree_id': 2
        },
    ]

    code, json = http.get('/api/fruit')
    assert code == 200
    assert json['occurences'] == 2
    assert idsorted(json['objects'], 'fruit_id') == [
        {
            'fruit_id': 1,
            'color': 'white',
            'size': 1.0,
            'age': 42.0,
            'tree_id': 1
        },
        {
            'fruit_id': 2,
            'color': 'white',
            'size': 2.0,
            'age': 42.0,
            'tree_id': 2
        },
    ]



from . import idsorted
from ..model import Fruit, Tree


def test_patch_tree(rest, http):
    rest(Tree, methods=['GET', 'PATCH'])
    code, json = http.patch('/api/tree/1', json={'id': 1, 'name': 'cedar'})
    assert code == 200
    assert json['occurences'] == 1
    assert idsorted(json['objects']) == [
        {
            'id': 1,
            'name': 'cedar'
        },
    ]

    code, json = http.get('/api/tree')
    assert code == 200
    assert json['occurences'] == 3
    assert idsorted(json['objects']) == [
        {
            'id': 1,
            'name': 'cedar'
        },
        {
            'id': 2,
            'name': 'maple'
        },
        {
            'id': 3,
            'name': 'oak'
        },
    ]


def test_patch_tree_without_id(rest, http):
    rest(Tree, methods=['GET', 'PATCH'])
    code, json = http.patch('/api/tree/1', json={'name': 'cedar'})
    assert code == 200
    assert json['occurences'] == 1
    assert idsorted(json['objects']) == [
        {
            'id': 1,
            'name': 'cedar'
        },
    ]

    code, json = http.get('/api/tree')
    assert code == 200
    assert json['occurences'] == 3
    assert idsorted(json['objects']) == [
        {
            'id': 1,
            'name': 'cedar'
        },
        {
            'id': 2,
            'name': 'maple'
        },
        {
            'id': 3,
            'name': 'oak'
        },
    ]


def test_patch_tree_with_another_id(rest, http):
    rest(Tree, methods=['GET', 'PATCH'])
    code, json = http.patch('/api/tree/1', json={'id': 5, 'name': 'cedar'})
    assert code == 500


def test_patch_fruit(rest, http):
    rest(Fruit, methods=['GET', 'PATCH'], allow_batch=True)
    code, json = http.patch(
        '/api/fruit/1', json={'fruit_id': 1,
                              'color': 'blue'}
    )
    assert code == 200
    assert json['occurences'] == 1
    assert idsorted(json['objects'], 'fruit_id') == [{
        'fruit_id': 1,
        'color': 'blue',
        'size': 12.0,
        'age': 1041300.0,
        'tree_id': 1
    }]

    code, json = http.get('/api/fruit')
    assert code == 200
    assert json['occurences'] == 5
    assert idsorted(json['objects'], 'fruit_id') == [{
        'fruit_id': 1,
        'color': 'blue',
        'size': 12.0,
        'age': 1041300.0,
        'tree_id': 1
    }, {
        'fruit_id': 2,
        'color': 'darkgrey',
        'size': 23.0,
        'age': 4233830.213,
        'tree_id': 1
    }, {
        'fruit_id': 3,
        'color': 'brown',
        'size': 2.12,
        'age': 0.0,
        'tree_id': 1
    }, {
        'fruit_id': 4,
        'color': 'red',
        'size': 0.5,
        'age': 2400.0,
        'tree_id': 2
    }, {
        'fruit_id': 5,
        'color': 'orangered',
        'size': 100.0,
        'age': 7200.000012,
        'tree_id': 2
    }]



from functools import wraps

from ..model import Tree


def test_auth_decorator(rest, http):
    def raise_if_id_is_2(fun):
        @wraps(fun)
        def wrapped(payload, id=None):
            if id == 2:
                rest.raise_error(403, 'is is 2')
            return fun(payload, id=id)

        return wrapped

    rest(Tree, methods=rest.all, auth=raise_if_id_is_2)
    assert http.get('/api/tree/1')[0] == 200
    assert http.get('/api/tree/2')[0] == 403
    assert http.put('/api/tree/1', json={'id': 1, 'name': 'cedar'})[0] == 200
    assert http.put('/api/tree/2', json={'id': 1, 'name': 'cedar'})[0] == 403
    assert http.post('/api/tree/1', json={'name': 'cedar'})[0] == 501
    assert http.post('/api/tree/2', json={'name': 'cedar'})[0] == 403
    assert http.delete('/api/tree/1')[0] == 200
    assert http.delete('/api/tree/2')[0] == 403


def test_read_write_decorator(rest, http):
    def raise_if_id_is_1(fun):
        @wraps(fun)
        def wrapped(payload, id=None):
            if id == 1:
                rest.raise_error(403, 'is is 1')
            return fun(payload, id=id)

        return wrapped

    def raise_if_id_is_2(fun):
        @wraps(fun)
        def wrapped(payload, id=None):
            if id == 2:
                rest.raise_error(403, 'is is 2')
            return fun(payload, id=id)

        return wrapped

    rest(
        Tree,
        methods=rest.all,
        read_auth=raise_if_id_is_1,
        write_auth=raise_if_id_is_2
    )
    assert http.get('/api/tree/1')[0] == 403
    assert http.get('/api/tree/2')[0] == 200
    assert http.put('/api/tree/1', json={'id': 1, 'name': 'cedar'})[0] == 200
    assert http.put('/api/tree/2', json={'id': 1, 'name': 'cedar'})[0] == 403
    assert http.post('/api/tree/1', json={'name': 'cedar'})[0] == 501
    assert http.post('/api/tree/2', json={'name': 'cedar'})[0] == 403
    assert http.delete('/api/tree/1')[0] == 200
    assert http.delete('/api/tree/2')[0] == 403


def test_all_decorator(rest, http):
    def raise_if_id_is_1(fun):
        @wraps(fun)
        def wrapped(payload, id=None):
            if id == 1:
                rest.raise_error(403, 'is is 1')
            return fun(payload, id=id)

        return wrapped

    def raise_if_id_is_2(fun):
        @wraps(fun)
        def wrapped(payload, id=None):
            if id == 2:
                rest.raise_error(403, 'is is 2')
            return fun(payload, id=id)

        return wrapped

    def raise_if_id_is_3(fun):
        @wraps(fun)
        def wrapped(payload, id=None):
            if id == 3:
                rest.raise_error(403, 'is is 3')
            return fun(payload, id=id)

        return wrapped

    rest(
        Tree,
        methods=rest.all,
        read_auth=raise_if_id_is_1,
        write_auth=raise_if_id_is_2,
        auth=raise_if_id_is_3
    )
    assert http.get('/api/tree/1')[0] == 403
    assert http.get('/api/tree/2')[0] == 200
    assert http.get('/api/tree/3')[0] == 403
    assert http.put('/api/tree/1', json={'id': 1, 'name': 'cedar'})[0] == 200
    assert http.put('/api/tree/2', json={'id': 1, 'name': 'cedar'})[0] == 403
    assert http.put('/api/tree/3', json={'id': 1, 'name': 'cedar'})[0] == 403
    assert http.post('/api/tree/1', json={'name': 'cedar'})[0] == 501
    assert http.post('/api/tree/2', json={'name': 'cedar'})[0] == 403
    assert http.post('/api/tree/3', json={'name': 'cedar'})[0] == 403
    assert http.delete('/api/tree/1')[0] == 200
    assert http.delete('/api/tree/2')[0] == 403
    assert http.delete('/api/tree/3')[0] == 403



from . import idsorted
from ..model import Tree


def test_post_tree_with_pk(rest, http):
    rest(Tree, methods=['GET', 'POST'])
    code, json = http.post('/api/tree/6', json={'name': 'mango'})
    assert code == 501


def test_post_tree_with_pk_custom(rest, http):
    tree = rest(Tree)

    @tree.declare('POST')
    def post(payload, id=None):
        # Make it work anyway
        return tree.post(payload)

    code, json = http.post('/api/tree/9', json={'name': 'mango'})
    assert code == 200
    assert json['occurences'] == 1
    assert idsorted(json['objects']) == [
        {
            'id': 4,
            'name': 'mango'
        },
    ]

    code, json = http.get('/api/tree')
    assert json['occurences'] == 4
    assert idsorted(json['objects']) == [
        {
            'id': 1,
            'name': 'pine'
        },
        {
            'id': 2,
            'name': 'maple'
        },
        {
            'id': 3,
            'name': 'oak'
        },
        {
            'id': 4,
            'name': 'mango'
        },
    ]



import json
import logging

from .__about__ import __uri__, __version__
from .coercers import Property
from .generators.openapi import OpenApi
from .generators.options import Options
from .rest import Rest

log = logging.getLogger('unrest')


class UnRest(object):
    """
    A troubling rest api library for sqlalchemy models.
    This is the main entry point of unrest.

    Common usage is as following:

    ```python
    rest = UnRest(app, session)  # app is your application
    # When called, it instanciate a `Rest` object that will register the
    # REST endpoints. See the `Rest` class.
    rest(Model1)
    rest(Model2)
    ```

    # Arguments
        app: Your web application,
            can be set afterwards using #UnRest.init_app
        session: Your sqlalchemy session,
            can be set afterwards using #UnRest.init_session
            and `init_session` method.
        path: Default '/api', sets the root url path for your endpoints
        version: Adds a version to the root url path if specified
            (i.e. /api/v2)
        framework: Your specific framework class, defaults to auto detect.
        SerializeClass: A global alternative for #Serialize class.
        DeserializeClass: A global alternative for #Deserialize class.
        allow_options: Set it to False to disable OPTIONS requests.
        serve_openapi_file: Set it to False to disable openapi file generation.
        info: Additional info for the openapi metadata.
    Unrest aims to be framework agnostic.
    It currently works with Flask out of the box, for another web framework
    you will have to implement your own Framework class.
    See `FlaskUnRest` in `flask_framework.py`
    """

    class RestError(Exception):
        """
        Exception raised by rest methods. It's catched by the REST method
        wrapper and will return a `status` http error with the specified
        `message`.
        """

        def __init__(self, status, message, extra=None):
            self.status = status
            self.message = message
            self.extra = extra or {}

    class ValidationError(Exception):
        """
        Exception raised by rest validation methods.
        """

        def __init__(self, message):
            self.message = message

    def __init__(
            self,
            app=None,
            session=None,
            path='/api',
            version='',
            framework=None,
            SerializeClass=None,
            DeserializeClass=None,
            allow_options=True,
            serve_openapi_file=True,
            openapi_class=OpenApi,
            options_class=Options,
            info={},
    ):
        self.rests = []
        self.path = path
        self.info = info
        self.version = version
        self._framework = framework
        self.SerializeClass = SerializeClass
        self.DeserializeClass = DeserializeClass
        self.allow_options = allow_options
        self.serve_openapi_file = serve_openapi_file
        self.OpenApi = openapi_class
        self.Options = options_class

        if app is not None:
            self.init_app(app)
        if session is not None:
            self.init_session(session)

    def init_app(self, app):
        """Sets the app on UnRest if it was missing during instantiation."""
        self.app = app
        prefix = self.root_path.lstrip('/').replace('/', '_')
        if self._framework:
            self.framework = self._framework(app, prefix=prefix)
        else:
            try:
                from flask import Flask
            except ImportError:
                pass
            else:
                if isinstance(app, Flask):
                    from .flask_framework import FlaskUnRest
                    self.framework = FlaskUnRest(app, prefix=prefix)
        if not self.framework:
            raise NotImplementedError(
                'Your framework %s is not recognized. '
                'Please provide a framework argument to UnRest' % type(app)
            )
        self.register_index()
        self.allow_options and self.register_options()
        self.serve_openapi_file and self.register_openapi()

    def init_session(self, session):
        """
        Sets the sqlalchemy session on UnRest
        if it was missing during instantiation.
        """
        self.session = session

    @property
    def root_path(self):
        if self.version:
            return '/'.join((self.path, self.version))
        return self.path

    @property
    def all(self):
        """
        Return all supported methods. Useful for the rest
        `method` keyword argument.
        """
        return ['GET', 'PUT', 'POST', 'DELETE', 'PATCH']

    def raise_error(self, status, message, extra=None):
        """
        Raise an error that will be handled by the rest wrapper, which
        will return a json response with status as HTTP status code
        and message as content.

        # Arguments
            status: The http status code corresponding to the error
                (404 for instance)
            message: The message that will be returned in the json response
            extra: Mapping of extra fields to return in json response
        """
        raise self.RestError(status, message, extra)

    def __call__(self, *args, **kwargs):
        """Returns a #unrest.Rest instance. See rest entry points."""

        if self.SerializeClass is not None:
            kwargs.setdefault('SerializeClass', self.SerializeClass)
        if self.DeserializeClass is not None:
            kwargs.setdefault('DeserializeClass', self.DeserializeClass)

        rest = Rest(self, *args, **kwargs)
        return rest

    def register_index(self):
        self.framework.register_route(
            self.root_path + '/', 'GET', None, self.index
        )

    def index(self):
        return (
            '<h1>unrest <small>api server</small></h1> version %s '
            '<a href="%s">unrest</a>'
        ) % (__version__, __uri__) + (
            ' <a href="%s/openapi.json">openapi.json</a>' % self.root_path
        ) if self.serve_openapi_file else ''

    def register_options(self):
        self.framework.register_route(
            self.root_path, 'OPTIONS', None, self.options
        )

    def options(self):
        return self.framework.send_json(json.dumps(self.Options(self).all()))

    def register_openapi(self):
        self.framework.register_route(
            self.root_path + '/openapi.json', 'GET', None, self.openapi
        )

    def openapi(self):
        return self.framework.send_json(json.dumps(self.OpenApi(self).all()))

    Property = Property

    class Response(object):
        def __init__(self, data, wrapper=lambda x: x, status_code=200):
            if isinstance(data, self.__class__):
                self.data = data.data
                self.wrapper = lambda r: wrapper(data.wrapper(r))
            else:
                self.data = data
                self.wrapper = wrapper
            self.status_code = status_code



from . import idsorted
from ..model import Fruit, Tree


def test_get_tree_with_relationship(rest, http):
    fruit = rest(Fruit, methods=[], only=['color'])

    rest(Tree, relationships={'fruits': fruit})
    code, json = http.get('/api/tree')
    assert code == 200
    assert json['occurences'] == 3
    assert idsorted(json['objects']) == [
        {
            'id':
                1,
            'name':
                'pine',
            'fruits': [{
                'fruit_id': 1,
                'color': 'grey'
            }, {
                'fruit_id': 2,
                'color': 'darkgrey'
            }, {
                'fruit_id': 3,
                'color': 'brown'
            }]
        },
        {
            'id':
                2,
            'name':
                'maple',
            'fruits': [{
                'fruit_id': 4,
                'color': 'red'
            }, {
                'fruit_id': 5,
                'color': 'orangered'
            }]
        },
        {
            'id': 3,
            'name': 'oak',
            'fruits': []
        },
    ]
    code, json = http.get('/api/fruit')
    assert code == 404


def test_get_fruit_with_relationship(rest, http):
    tree = rest(Tree, methods=[])
    rest(Fruit, only=['color'], relationships={'tree': tree})

    code, json = http.get('/api/fruit')
    assert code == 200
    assert json['occurences'] == 5
    assert idsorted(json['objects'], 'fruit_id') == [{
        'fruit_id': 1,
        'color': 'grey',
        'tree': [{
            'id': 1,
            'name': 'pine'
        }]
    }, {
        'fruit_id': 2,
        'color': 'darkgrey',
        'tree': [{
            'id': 1,
            'name': 'pine'
        }]
    }, {
        'fruit_id': 3,
        'color': 'brown',
        'tree': [{
            'id': 1,
            'name': 'pine'
        }]
    }, {
        'fruit_id': 4,
        'color': 'red',
        'tree': [{
            'id': 2,
            'name': 'maple'
        }]
    }, {
        'fruit_id': 5,
        'color': 'orangered',
        'tree': [{
            'id': 2,
            'name': 'maple'
        }]
    }]
    code, json = http.get('/api/tree')
    assert code == 404



import logging

from flask import current_app, jsonify, request, url_for

log = logging.getLogger('unrest.flask')


class FlaskUnRest(object):
    """
    Unrest flask framework implementation.
    This is the framework abstraction you can implement for your own framework

    """

    def __init__(self, app, prefix):
        self.app = app
        self.prefix = prefix

    def register_route(self, path, method, parameters, fun):
        """
        Register the given function for `path` and `method` with and without
        `parameters`.

        # Arguments
            path: The url of the endoint without arguments. ('/api/person')
            method: The HTTP method to register the route on.
            parameters: The primary keys of the model that can be given
                after the path. `PrimaryKey('id'), PrimaryKey('type')) -> \
'/api/person/<id>/<type>'`
            fun: The route function
        """
        name = self._name(fun.__name__)

        if self.app.view_functions.pop(name, None):
            log.info('Overriding route %s' % name)
        getattr(fun, '__func__', fun).provide_automatic_options = False
        self.app.add_url_rule(path, name, fun, methods=[method])
        if parameters:
            log.info(
                'Registering route %s for %s for %s' % (name, path, method)
            )
            path_with_params = path + '/' + '/'.join(
                '<%s>' % param for param in parameters
            )

            log.info(
                'Registering route for %s for %s' % (path_with_params, method)
            )
            self.app.add_url_rule(
                path_with_params, name, fun, methods=[method]
            )

    def request_json(self):
        """
        Must return the string of the current JSON request content or None
        """
        if not request.is_json:
            return None

        return request.data.decode('utf-8')

    def send_json(self, json, status_code=200):
        """
        Send a `status_code` JSON response with `json` as content

        # Arguments
            json: The JSON string to send.
            status_code: The response status code. (Default: 200)
        """
        return current_app.response_class(
            (json, '\n'),
            status=status_code,
            mimetype=current_app.config['JSONIFY_MIMETYPE']
        )

    def send_error(self, message, status_code):
        """
        Send an error as a JSON response with the given status code.

        # Arguments
            message: The JSON string containing the error message.
            status_code: The HTTP status code (i.e. 402)
        """
        return jsonify(message), status_code

    @property
    def url(self):
        """
        Return the api url root
        """
        return url_for(self._name('index'), _external=True)

    def _name(self, name):
        """Generate a unique name for endpoint"""
        return 'unrest__%s__%s' % (self.prefix, name)
